from concourse.model.traits.notifications import NotificationTriggeringPolicy
from concourse.client.model import BuildStatus

import ccc.concourse
import concourse.util

import os
import traceback

import ci.util
import mailutil


def meta_vars():
    build = concourse.util.find_own_running_build()
    pipeline_metadata = concourse.util.get_pipeline_metadata()
    config_set = ci.util.ctx().cfg_factory().cfg_set(pipeline_metadata.current_config_set_name)
    concourse_cfg = config_set.concourse()
    v = {
        'build-id': build.id(),
        'build-name': build.build_number(),
        'build-job-name': pipeline_metadata.job_name,
        'build-team-name': pipeline_metadata.team_name,
        'build-pipeline-name': pipeline_metadata.pipeline_name,
        'atc-external-url': concourse_cfg.external_url(),
    }

    return v


def job_url(v):
    return '/'.join([
        v['atc-external-url'],
        'teams',
        v['build-team-name'],
        'pipelines',
        v['build-pipeline-name'],
        'jobs',
        v['build-job-name'],
        'builds',
        v['build-name']
    ])


def determine_previous_build_status(v, cfg_set):
    cc_cfg = cfg_set.concourse()

    concourse_client = ccc.concourse.client_from_cfg_name(
        concourse_cfg_name=cc_cfg.name(),
        team_name=v['build-team-name']
    )
    try:
        build_number = int(float(v['build-name']))
        if build_number < 2:
            ci.util.info('this seems to be the first build - will notify')
            return BuildStatus.SUCCEEDED

        previous_build = str(build_number - 1)
        previous_build = concourse_client.job_build(
            pipeline_name=v['build-pipeline-name'],
            job_name=v['build-job-name'],
            build_name=previous_build
        )
        return previous_build.status()
    except Exception as e:
        if type(e) == SystemExit:
            raise e
        # in doubt, ensure notification is sent
        traceback.print_exc()
        return None


def should_notify(
    triggering_policy,
    meta_vars,
    cfg_set,
    determine_previous_build_status=determine_previous_build_status,
):
    if triggering_policy == NotificationTriggeringPolicy.ALWAYS:
        return True
    elif triggering_policy == NotificationTriggeringPolicy.NEVER:
        return False
    elif triggering_policy == NotificationTriggeringPolicy.ONLY_FIRST:
        previous_build_status = determine_previous_build_status(meta_vars, cfg_set)
        if not previous_build_status:
            ci.util.info('failed to determine previous build status - will notify')
            return True

        # assumption: current job failed
        if previous_build_status in (BuildStatus.FAILED, BuildStatus.ERRORED):
          ci.util.info('previous build was already broken - will not notify')
          return False
        return True
    else:
        raise NotImplementedError


def cfg_from_callback(
    repo_root,
    callback_path,
    effective_cfg_file,
):
    import os
    import subprocess
    import tempfile

    import ci.util

    tmp_file = tempfile.NamedTemporaryFile()
    cb_env = os.environ.copy()
    cb_env['REPO_ROOT'] = repo_root
    cb_env['NOTIFY_CFG_OUT'] = tmp_file.name
    cb_env['EFFECTIVE_CFG'] = effective_cfg_file

    subprocess.run(
        [callback_path],
        check=True,
        env=cb_env,
    )

    return ci.util.parse_yaml_file(tmp_file.name)


def components_with_version_changes(component_diff_path: str):
    if not os.path.isfile(component_diff_path):
        ci.util.info('no component_diff found at: ' + str(component_diff_path))
        return set()
    else:
        component_diff = ci.util.parse_yaml_file(component_diff_path)
        comp_names = component_diff.get('component_names_with_version_changes', set())
        return set(comp_names)


def retrieve_build_log(concourse_api, task_name):
    v = meta_vars()
    try:
      build_id = v['build-id']
      task_id = concourse_api.build_plan(build_id=build_id).task_id(task_name=task_name)
      build_events = concourse_api.build_events(build_id=build_id)
      build_log = '\n'.join(build_events.iter_buildlog(task_id=task_id))
      return build_log
    except Exception:
      traceback.print_exc() # print_err, but send email notification anyway
      return 'failed to retrieve build log'


def resolve_recipients_by_component(components, github_cfg_name):
    component_recipients = mailutil.determine_mail_recipients(
        github_cfg_name=github_cfg_name, # todo: actually this is not required here
        components=components,
    )
    return component_recipients



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import logging

import ci.log
from model.base import (
    NamedModelElement,
)
import model.concourse
import model.secret


logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


class SecretsServerConfig(NamedModelElement):
    def _required_attributes(self):
        return {
            'namespace',
            'service_name',
        }

    def _optional_attributes(self):
        return {
            'node_selector',
        }

    def namespace(self):
        return self.raw.get('namespace')

    def service_name(self):
        return self.raw.get('service_name')

    def endpoint_url(self):
        return f'http://{self.service_name()}.{self.namespace()}.svc.cluster.local'

    def node_selector(self):
        return self.raw.get('node_selector')


def secret_url_path(
    job_mapping: model.concourse.JobMapping,
    secret_cfg: model.secret.Secret,
):
    '''
        used to retrieve the secret url path for given config in default template
    '''
    if not secret_cfg:
        logger.warning(f'No secret configuration found, check job_mapping {job_mapping.name()}')
        return

    if job_mapping.secrets_repo():
        if secret_cfg.generation():
            return _org_based_secret_url_path_with_generation(
                target_secret_name=job_mapping.target_secret_name(),
                secret_cfg_name=job_mapping.target_secret_cfg_name(),
                generation=secret_cfg.generation(),
            )
        else:
            return _org_based_secret_url_path(
                target_secret_name=job_mapping.target_secret_name(),
                secret_cfg_name=job_mapping.target_secret_cfg_name(),
            )
    else:
        logger.warning(
            f'No secrets repo for job_mapping {job_mapping.name()} configured. Please do so...',
        )


def _org_based_secret_url_path(target_secret_name, secret_cfg_name):
    return f'{target_secret_name}/{secret_cfg_name}'


def _org_based_secret_url_path_with_generation(target_secret_name, secret_cfg_name, generation):
    return f'{target_secret_name}-{generation}/{secret_cfg_name}'



# SPDX-FileCopyrightText: 2022 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import re
import typing

import deprecated

import oci.model as om


class PlatformFilter:
    def __init__(
        self,
        included_platform_regexes: typing.Iterable[str],
    ):
        '''
        @param included_platforms: regexes to filter for platforms in format os/arch/variant
        '''
        self.included_platform_regexes = tuple(included_platform_regexes)

    def __call__(self, platform: om.OciPlatform) -> bool:
        '''
        returns True if the passed platform matches this filter (i.e. should be inclued), else False
        '''
        normalised_platform = normalise(platform)
        for platform_regex in self.included_platform_regexes:
            if re.fullmatch(platform_regex, normalised_platform):
                return True

        return False

    def __repr__(self):
        return f'{self.__class__.__name__}({self.included_platform_regexes=})'

    @deprecated.deprecated
    @staticmethod
    def create(
        included_platforms: typing.List[str],
    ) -> typing.Callable[[om.OciPlatform], bool]:
        return PlatformFilter(included_platform_regexes=included_platforms)


def normalise(p: om.OciPlatform):
    os = normalise_os(p.os)
    arch, variant = normalise_arch(p.architecture, p.variant)

    normalised = os + '/' + arch
    if not variant == '':
        normalised += '/' + variant

    return normalised


def normalise_os(os: str) -> str:
    '''
    https://github.com/containerd/containerd/blob/8686ededfc90076914c5238eb96c883ea093a8ba/platforms/database.go#L69
    '''
    if not os:
        raise ValueError(os)

    os = os.lower()
    if os == 'macos':
        os = 'darwin'

    return os


def normalise_arch(arch: str, variant: str) -> typing.Tuple:
    '''
    https://github.com/containerd/containerd/blob/8686ededfc90076914c5238eb96c883ea093a8ba/platforms/database.go#L83
    '''
    if not arch:
        raise ValueError(arch)

    variant = variant or ''
    arch, variant = arch.lower(), variant.lower()
    match arch:
        case 'i386':
            arch = '386'
            variant = ''
        case 'x86_64', 'x86-64':
            arch = 'amd64'
            variant = ''
        case 'aarch64', 'arm64':
            arch = 'arm64'
            if variant in ('8', 'v8'):
                variant = ''
        case 'armhf':
            arch = 'arm'
            variant = 'v7'
        case 'armel':
            arch = 'arm'
            variant = 'v6'
        case 'arm':
            if variant in ('', '7'):
                variant = 'v7'
            elif variant in ('5', '6', '8'):
                variant = 'v' + variant

    return arch, variant



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    ModelBase,
    NamedModelElement,
)

import github.util

from github3 import GitHub

TEKTON_GITHUB_ORG = 'tektoncd'
TEKTON_PIPELINES_REPO_NAME = 'pipeline'
TEKTON_DASHBOARD_REPO_NAME = 'dashboard'

TEKTON_PIPELINES_RELEASE_ASSET_NAME = 'release.yaml'
TEKTON_DASHBOARD_RELEASE_ASSET_NAME = 'tekton-dashboard-release.yaml'


class TektonConfig(NamedModelElement):
    '''Not intended to be instantiated by users of this module
    '''
    def pipelines_config(self):
        if raw_cfg := self.raw.get('pipelines'):
            return TektonPipelinesConfig(raw_cfg)
        return None

    def dashboard_config(self):
        if raw_cfg := self.raw.get('dashboard'):
            return TektonDashboardConfig(raw_cfg)
        return None

    def kubernetes_config_name(self):
        return self.raw['kubernetes_config']

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'kubernetes_config',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'dashboard',
            'pipelines',
        ]


class TektonPipelinesConfig(ModelBase):
    def namespace(self):
        return self.raw.get('namespace')

    def version(self):
        return self.raw.get('version')

    def install_manifests(self):
        gh_helper = github.util.GitHubRepositoryHelper(
            owner=TEKTON_GITHUB_ORG,
            name=TEKTON_PIPELINES_REPO_NAME,
            github_api=GitHub(),
        )
        return gh_helper.retrieve_asset_contents(
            release_tag=self.version(),
            asset_label=TEKTON_PIPELINES_RELEASE_ASSET_NAME,
        )

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'namespace'
            'version',
        ]


class TektonDashboardConfig(ModelBase):
    def namespace(self):
        return self.raw.get('namespace')

    def version(self):
        return self.raw.get('version')

    def install_manifests(self):
        gh_helper = github.util.GitHubRepositoryHelper(
            owner=TEKTON_GITHUB_ORG,
            name=TEKTON_DASHBOARD_REPO_NAME,
            github_api=GitHub(),
        )
        return gh_helper.retrieve_asset_contents(
            release_tag=self.version(),
            asset_label=TEKTON_DASHBOARD_RELEASE_ASSET_NAME,
        )

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'namespace'
            'version',
        ]



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import unittest

import shlex

from concourse.model.base import ScriptType
from concourse.model.step import PipelineStep


class PipelineStepTest(unittest.TestCase):
    def _examinee(self, name='dontcare',  **kwargs):
        return PipelineStep(
            name=name,
            is_synthetic=False,
            script_type=ScriptType.BOURNE_SHELL,
            raw_dict=kwargs,
        )

    def test_image(self):
        examinee = self._examinee(image='an_image:1.2.3')
        self.assertEqual(examinee.image(), 'an_image:1.2.3')

    def test__argv(self):
        # argv defaults to [step.name]
        examinee = self._examinee(name='a_name')
        self.assertEqual(examinee._argv(), ['a_name'])

        # executable may be overwritten
        examinee = self._examinee(execute='another_executable')
        self.assertEqual(examinee._argv(), ['another_executable'])

        # executable may be a list
        examinee = self._examinee(execute=['a', 'b'])
        self.assertEqual(examinee._argv(), ['a', 'b'])

    def test_executable(self):
        examinee = self._examinee(name='x')
        self.assertEqual(examinee.executable(), 'x')
        self.assertEqual(examinee.executable(prefix='foo'), 'foo/x')
        self.assertEqual(examinee.executable(prefix=('foo',)), 'foo/x')
        self.assertEqual(examinee.executable(prefix=('foo','bar')), 'foo/bar/x')

        examinee = self._examinee(execute='another_executable')
        self.assertEqual(examinee.executable(), 'another_executable')

        examinee = self._examinee(execute=['exec', 'arg 1', 'arg2'])
        self.assertEqual(examinee.executable(), 'exec')

    def test_execute(self):
        examinee = self._examinee(execute=['exec', 'arg1'])
        self.assertEqual(examinee.execute(), 'exec arg1')

        # whitespace must be quoted
        examinee = self._examinee(execute=['e x', 'a r g'])
        self.assertEqual(examinee.execute(), ' '.join(map(shlex.quote, ('e x', 'a r g'))))



import os
import pytest

import git

import gitutil

import concourse.steps.release as release_step
import concourse.model.traits.version as version_trait


@pytest.fixture
def git_helper(tmpdir):
    print(tmpdir)
    repo = git.Repo.init(tmpdir)
    repo.index.commit(message='first (empty) commit')

    git_helper = gitutil.GitHelper(repo=repo, github_cfg=None, github_repo_path='org/repo')

    return git_helper


def test_create_release_commit(git_helper):
    work_tree = git_helper.repo.working_tree_dir
    version_file = os.path.join(work_tree, 'version.txt')
    with open(version_file, 'w') as f:
        f.write('dummy')
    git_helper.add_and_commit('add version')

    release_commit = release_step.create_release_commit(
        git_helper=git_helper,
        branch='master',
        version='1.2.3',
        version_interface=version_trait.VersionInterface.FILE,
        version_path=version_file,
        release_commit_message_prefix='my prefix',
    )

    repo = git_helper.repo

    # check returned commit is contained in repository (-> was created)
    assert 'my prefix' in release_commit.message

    # release-commit should not be visible in working tree
    with open(version_file) as f:
        assert f.read() == 'dummy'

    # check commit actually contains correct (versionfile) diff
    repo.head.reset(release_commit, working_tree=True)
    with open(version_file) as f:
        assert f.read() == '1.2.3'



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import datetime
import logging

import ccc.elasticsearch
from model.webhook_dispatcher import WebhookDispatcherConfig
from .dispatcher import GithubWebhookDispatcher
from .model import CreateEvent, PushEvent, PullRequestEvent

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class GithubWebhook:
    def __init__(
        self,
        cfg_factory,
        cfg_set,
        whd_cfg: WebhookDispatcherConfig,
        es_client: ccc.elasticsearch.ElasticSearchClient,
    ):
        self.dispatcher = GithubWebhookDispatcher(
            cfg_factory=cfg_factory,
            cfg_set=cfg_set,
            whd_cfg=whd_cfg,
        )
        self.es_client = es_client

    def on_post(self, req, resp):
        event = req.get_header('X-GitHub-Event', required=True)
        delivery = req.get_header('X-GitHub-Delivery', required=True)
        logger_string = f'received event (delivery-id: {delivery}) of type "{event}"'
        action = req.media.get("action")
        if action:
            logger_string += f' with action "{action}"'
        repository_name = req.media.get("repository", {}).get("full_name")
        # no GHES header ^= github.com
        hostname = req.get_header('X-GitHub-Enterprise-Host') or 'github.com'
        if repository_name:
            logger_string += f' for repository "{repository_name}"'
        if hostname:
            logger_string += f' from "{hostname}"'

        logger.info(logger_string)
        dispatch_start_time = datetime.datetime.now()
        if event == 'push':
            parsed = PushEvent(raw_dict=req.media, delivery=delivery, hostname=hostname)
            self.dispatcher.dispatch_push_event(
                push_event=parsed,
                es_client=self.es_client,
                delivery_id=delivery,
                hostname=hostname,
                repository=repository_name,
                dispatch_start_time=dispatch_start_time,
            )
            return
        if event == 'create':
            parsed = CreateEvent(raw_dict=req.media, delivery=delivery, hostname=hostname)
            self.dispatcher.dispatch_create_event(
                create_event=parsed,
                es_client=self.es_client,
                delivery_id=delivery,
                hostname=hostname,
                repository=repository_name,
                dispatch_start_time=dispatch_start_time,
            )
            return
        elif event == 'pull_request':
            parsed = PullRequestEvent(raw_dict=req.media, delivery=delivery, hostname=hostname)
            processing = self.dispatcher.dispatch_pullrequest_event(
                pr_event=parsed,
                es_client=self.es_client,
                dispatch_start_time=dispatch_start_time,
            )
            if not processing:
                resp.text = "Event ignored"
            return
        else:
            msg = f'event {event} ignored'
            logger.info(msg)
            return



import dataclasses
import datetime
import json
import typing


@dataclasses.dataclass(frozen=True)
class ExceptionMetric:
    service: str
    stacktrace: typing.List[str]
    request: str
    params: str
    creation_date: str

    @staticmethod
    def create(
        service: str,
        stacktrace: typing.List[str],
        request: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
    ) -> 'ExceptionMetric':
        '''
        convenience method to create a `ExceptionMetric`
        '''
        return ExceptionMetric(
            creation_date=datetime.datetime.now().isoformat(),
            service=service,
            stacktrace=stacktrace,
            request=json.dumps(request),
            params=json.dumps(params),
        )


@dataclasses.dataclass(frozen=True)
class WebhookDelivery:
    deliveryId: str
    eventType: str #TODO: introduce enum
    repository: str
    hostname: str
    processTotalSeconds: float
    creation_date: str

    @staticmethod
    def create(
        delivery_id: str,
        event_type: str,
        repository: str,
        hostname: str,
        process_total_seconds: float
    ) -> 'WebhookDelivery':
        '''
        convenience method to create a `WebhookDelivery`
        '''
        return WebhookDelivery(
            creation_date=datetime.datetime.now().isoformat(),
            deliveryId=delivery_id,
            eventType=event_type,
            repository=repository,
            hostname=hostname,
            processTotalSeconds=process_total_seconds,
        )


@dataclasses.dataclass(frozen=True)
class WebhookResourceUpdateFailed:
    deliveryId: str
    repository: str
    hostname: str
    eventType: str
    outdatedResourcesNames: typing.List[str]
    prId: int
    prAction: str
    creation_date: str

    @staticmethod
    def create(
        delivery_id: str,
        repository: str,
        hostname: str,
        event_type: str,
        outdated_resources_names: typing.List[str],
        pr_id: int,
        pr_action: str,
    ) -> 'WebhookResourceUpdateFailed':
        '''
        convenience method to create a `WebhookResourceUpdateFailed`
        '''
        return WebhookResourceUpdateFailed(
            creation_date=datetime.datetime.now().isoformat(),
            deliveryId=delivery_id,
            repository=repository,
            hostname=hostname,
            eventType=event_type,
            outdatedResourcesNames=outdated_resources_names,
            prId=pr_id,
            prAction=pr_action,
        )


def index_name(
    obj: typing.Union[WebhookDelivery, WebhookResourceUpdateFailed, ExceptionMetric],
) -> str:
    if isinstance(obj, WebhookDelivery):
        return 'webhook_delivery'
    elif isinstance(obj, WebhookResourceUpdateFailed):
        return 'webhook_resource_update_failed'
    elif isinstance(obj, ExceptionMetric):
        return 'cicd_services_exception'

    raise NotImplementedError(obj)



import dataclasses
import logging
import os
import subprocess
import tempfile
import typing
import version

import yaml

from github3.exceptions import (
    ConnectionError,
)
from git.exc import (
    GitCommandError
)
import git.types

import ocm

import concourse.steps.version
import concourse.model.traits.version as version_trait
import dockerutil
import release_notes.fetch
import release_notes.markdown
import slackclient.util

from gitutil import GitHelper
from github.util import (
    GitHubRepositoryHelper,
)
from concourse.model.traits.release import (
    ReleaseCommitPublishingPolicy,
)
import model.container_registry as cr

logger = logging.getLogger('step.release')


def _invoke_callback(
    callback_script_path: str,
    repo_dir: str,
    effective_version: str,
    callback_image_reference: str=None,
):
    callback_env = os.environ.copy()
    callback_env['EFFECTIVE_VERSION'] = effective_version

    if callback_image_reference:
        repo_dir_in_container = '/mnt/main_repo'
        callback_env['REPO_DIR'] = repo_dir_in_container
    else:
        callback_env['REPO_DIR'] = repo_dir

    if not callback_image_reference:
        callback_script_path = os.path.join(
            repo_dir,
            callback_script_path,
        )
        subprocess.run(
            [callback_script_path],
            check=True,
            env=callback_env,
        )
    else:
        script_path_in_container = os.path.join(
            repo_dir_in_container,
            callback_script_path,
        )

        oci_registry_cfg = cr.find_config(image_reference=callback_image_reference)
        if oci_registry_cfg:
            docker_cfg_dir = tempfile.TemporaryDirectory()
            dockerutil.mk_docker_cfg_dir(
                cfg={'auths': oci_registry_cfg.as_docker_auths()},
                cfg_dir=docker_cfg_dir.name,
                exist_ok=True,
            )
        else:
            docker_cfg_dir = None

        docker_argv = dockerutil.docker_run_argv(
            image_reference=callback_image_reference,
            argv=(script_path_in_container,),
            env=callback_env,
            mounts={
                repo_dir: repo_dir_in_container,
            },
            cfg_dir=docker_cfg_dir.name,
        )

        dockerutil.launch_dockerd_if_not_running()

        logger.info(f'will run callback using {docker_argv=}')

        try:
            subprocess.run(
                docker_argv,
                check=True,
            )
        finally:
            if docker_cfg_dir:
                docker_cfg_dir.cleanup()


def _calculate_next_cycle_dev_version(
    release_version: str,
    version_operation: str,
    prerelease_suffix: str,
):
    # calculate the next version and append the prerelease suffix
    return version.process_version(
        version_str=version.process_version(
            version_str=release_version,
            operation=version_operation,
        ),
        operation='set_prerelease',
        prerelease=prerelease_suffix,
    )


def _calculate_tags(
    version: str,
    github_release_tag: dict,
    git_tags: list,
) -> typing.Sequence[str]:
    github_release_tag_candidate = github_release_tag['ref_template'].format(
        VERSION=version,
    )
    git_tag_candidates = [
        tag_template['ref_template'].format(VERSION=version)
        for tag_template in git_tags
    ]

    return [github_release_tag_candidate] + git_tag_candidates


def collect_release_notes(
    repo_dir,
    release_version: str,
    component,
    component_descriptor_lookup,
    version_lookup,
) -> str:
    release_note_blocks = release_notes.fetch.fetch_release_notes(
        repo_path=repo_dir,
        component=component,
        version_lookup=version_lookup,
        component_descriptor_lookup=component_descriptor_lookup,
        current_version=release_version,
    )

    release_notes_markdown = '\n'.join(
        str(i) for i in release_notes.markdown.render(release_note_blocks)
    ) or 'no release notes available'

    if (component_resources_markdown := release_notes.markdown.release_note_for_ocm_component(
        component=component,
    )):
        release_notes_markdown += '\n\n' + component_resources_markdown

    return release_notes_markdown


def have_tag_conflicts(
    github_helper,
    tags,
):
    found_tags = 0
    for tag in tags:
        if github_helper.tag_exists(tag.removeprefix('refs/tags/')):
            logger.error(f'{tag=} exists in remote repository - aborting release')
            found_tags += 1

    if not found_tags:
        return False

    logger.error('HINT: manually bump VERSION or remove tag')
    return True


def create_release_commit(
    git_helper,
    branch: str,
    version: str,
    version_interface,
    version_path: str,
    release_commit_message_prefix: str='',
    release_commit_callback: str=None,
    release_commit_callback_image_reference: str=None,
) -> git.Commit:
    # clean repository if required
    worktree_dirty = bool(git_helper._changed_file_paths())
    if worktree_dirty:
        git_helper.repo.head.reset(working_tree=True)

    commit_message = f'Release {version}'
    if release_commit_message_prefix:
        commit_message = f'{release_commit_message_prefix} {commit_message}'

    concourse.steps.version.write_version(
        version_interface=version_interface,
        version_str=version,
        path=version_path,
    )

    if release_commit_callback:
        _invoke_callback(
            callback_script_path=release_commit_callback,
            repo_dir=git_helper.repo.working_tree_dir,
            effective_version=version,
            callback_image_reference=release_commit_callback_image_reference,
        )

    release_commit = git_helper.index_to_commit(
        message=commit_message,
    )

    # make sure working tree is clean for later git operations
    if git_helper._changed_file_paths():
        git_helper.repo.head.reset(index=True, working_tree=True)

    return release_commit


def create_and_push_bump_commit(
    git_helper: GitHelper,
    repo_dir: str,
    release_commit: git.Commit,
    merge_release_back_to_default_branch_commit: git.Commit,
    release_version: str,
    version_interface: version_trait.VersionInterface,
    version_path: str,
    repository_branch: str,
    version_operation: str,
    prerelease_suffix: str,
    publishing_policy: ReleaseCommitPublishingPolicy,
    commit_message_prefix: str='',
    next_version_callback: str=None,
):
    # clean repository if required
    worktree_dirty = bool(git_helper._changed_file_paths())

    if worktree_dirty:
        if publishing_policy is ReleaseCommitPublishingPolicy.TAG_AND_PUSH_TO_BRANCH:
            reset_to = release_commit.hexsha
        elif publishing_policy is ReleaseCommitPublishingPolicy.TAG_ONLY:
            reset_to = 'HEAD'
        elif publishing_policy is ReleaseCommitPublishingPolicy.TAG_AND_MERGE_BACK:
            reset_to = merge_release_back_to_default_branch_commit or 'HEAD'
        else:
            raise NotImplementedError

        git_helper.repo.head.reset(
            commit=reset_to,
            index=True,
            working_tree=True,
        )

    # reset to release-commit so potential changes from release-commit-callback are not
    # unintentionally reverted
    if publishing_policy is ReleaseCommitPublishingPolicy.TAG_AND_PUSH_TO_BRANCH:
        git_helper.repo.head.reset(
            commit=release_commit,
            index=True,
            working_tree=True,
        )

    # prepare next dev cycle commit
    next_version = _calculate_next_cycle_dev_version(
        release_version=release_version,
        version_operation=version_operation,
        prerelease_suffix=prerelease_suffix,
    )
    logger.info(f'{next_version=}')

    concourse.steps.version.write_version(
        version_interface=version_interface,
        version_str=next_version,
        path=version_path,
    )

    # call optional dev cycle commit callback
    if next_version_callback:
        _invoke_callback(
            callback_script_path=next_version_callback,
            repo_dir=repo_dir,
            effective_version=next_version,
        )

    # depending on publishing-policy, bump-commit should become successor of
    # either the release commit, or just be pushed to branch-head
    if publishing_policy is ReleaseCommitPublishingPolicy.TAG_AND_PUSH_TO_BRANCH:
        parent_commits = [release_commit.hexsha]
    elif publishing_policy is ReleaseCommitPublishingPolicy.TAG_ONLY:
        parent_commits = None # default to current branch head
    elif publishing_policy is ReleaseCommitPublishingPolicy.TAG_AND_MERGE_BACK:
        parent_commit = git_helper.repo.head.commit

        if parent_commit:
            parent_commits = [parent_commit]
        else:
            parent_commits = None # default to current branch head

    commit_message = f'Prepare next Development Cycle {next_version}'
    if commit_message_prefix:
        commit_message = f'{commit_message_prefix} {commit_message}'

    next_cycle_commit = git_helper.index_to_commit(
        message=commit_message,
        parent_commits=parent_commits,
    )

    git_helper.push(
        from_ref=next_cycle_commit.hexsha,
        to_ref=repository_branch,
    )


def create_and_push_tags(
    git_helper,
    tags,
    release_commit: git.Commit,
):
    for tag in tags:
        try:
            git_helper.push(
                from_ref=release_commit.hexsha,
                to_ref=tag,
            )
        except GitCommandError:
            logger.error(f'failed to push {tag=}')
            raise


def create_and_push_mergeback_commit(
    git_helper,
    github_helper,
    tags,
    branch: str,
    merge_commit_message_prefix: str,
    release_commit: git.Commit,
):
    release_tag = tags[0]
    commit_message = f'Merge release-commit from {release_tag} into {branch}'
    if merge_commit_message_prefix:
        commit_message = f'{merge_commit_message_prefix} {commit_message}'

    git_repo = git_helper.repo

    git_repo.head.reset(
        commit=release_commit,
        index=True,
        working_tree=True,
    )

    # fetch and rebase (again, in case there was a head-update)
    upstream_commit = git_helper.fetch_head(
        f'refs/heads/{branch}'
    )
    git_helper.rebase(commit_ish=upstream_commit.hexsha)

    # update submodules (if any), to avoid local diff to be included in subsequent `git add`.
    # otherwise, upstream submodule-updates might be reverted by us.
    git_helper.submodule_update()

    # create merge commit
    git_repo.index.merge_tree(
        release_commit,
        git_repo.merge_base(upstream_commit, release_commit),
    )
    merge_commit = git_helper.index_to_commit(
        message=commit_message,
        parent_commits=(
            upstream_commit,
            release_commit,
        ),
    )

    git_helper.push(
        from_ref=merge_commit.hexsha,
        to_ref=branch
    )

    git_repo.head.reset(
        commit=merge_commit.hexsha,
        index=True,
        working_tree=True,
    ) # make sure next dev-cycle commit does not undo the merge-commit


def github_release(
    github_helper: GitHubRepositoryHelper,
    release_tag: str,
    release_version: str,
    component_name: str,
):
    # github-api expects unqualified tagname
    release_tag = release_tag.removeprefix('refs/tags/')

    if release := github_helper.draft_release_with_name(f'{release_version}-draft'):
        github_helper.promote_draft_release(
            draft_release=release,
            release_tag=release_tag,
            release_version=release_version,
            component_name=component_name,
        )
    else:
        release = github_helper.create_release(
            tag_name=release_tag,
            body='',
            draft=False,
            prerelease=False,
            name=release_version,
            component_name=component_name,
        )


def upload_github_release_asset(
    github_helper: GitHubRepositoryHelper,
    github_release_tag: str,
    component,
):
    # upload copy as release-asset
    release_tag_name = github_release_tag.removeprefix('refs/tags/')
    try:
        release = github_helper.repository.release_from_tag(release_tag_name)

        component_descriptor = ocm.ComponentDescriptor(
            component=component,
            meta=ocm.Metadata(),
            signatures=[],
        )

        descriptor_str = yaml.dump(
            data=dataclasses.asdict(component_descriptor),
            Dumper=ocm.EnumValueYamlDumper,
        )

        normalized_component_name = component.name.replace('/', '_')
        asset_name = f'{normalized_component_name}.component_descriptor.cnudie.yaml'
        release.upload_asset(
            content_type='application/x-yaml',
            name=asset_name,
            asset=descriptor_str.encode('utf-8'),
            label=asset_name,
        )
    except ConnectionError:
        logger.warning('Unable to attach component-descriptors to release as release-asset.')


def clean_draft_releases(
    github_helper: GitHubRepositoryHelper,
):
    for release, deletion_successful in github_helper.delete_outdated_draft_releases():
        if deletion_successful:
            logger.info(f'Deleted draft {release.name=}')
        else:
            logger.warning(f'Could not delete draft {release.name=}')


def post_to_slack(
    release_notes_markdown,
    component: ocm.Component,
    slack_cfg_name: str,
    slack_channel: str,
):
    responses = slackclient.util.post_to_slack(
        release_notes_markdown=release_notes_markdown,
        component_name=component.name,
        release_version=component.version,
        slack_cfg_name=slack_cfg_name,
        slack_channel=slack_channel,
    )

    for response in responses:
        if response and response.get('file', None):
            uploaded_file_id = response.get('file').get('id')
            logger.info(f'uploaded {uploaded_file_id=} to slack')
        else:
            raise RuntimeError('Unable to get file id from Slack response')
    logger.info('successfully posted contents to slack')



import ocm

import ioutil


def s3_access_as_blob_descriptor(
    s3_client: 'botocore.client.S3',
    s3_access: ocm.S3Access,
    chunk_size: int=4096,
    name: str=None,
) -> ioutil.BlobDescriptor:
    if not s3_client:
        raise ValueError('must pass-in s3-client')

    blob = s3_client.get_object(Bucket=s3_access.bucketName, Key=s3_access.objectKey)

    size = blob['ContentLength']
    body = blob['Body']

    return ioutil.BlobDescriptor(
        content=body.iter_chunks(chunk_size=chunk_size),
        size=size,
        name=name or f's3://{s3_access.bucketName}/{s3_access.objectKey}',
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import semver
import pytest

import version


def test_find_latest_version():
    versions = (semver.VersionInfo.parse(v) for v in (
            '0.0.10',
            '0.20.1',
            '2.50.100',
            '3.0.1',
            '1.0.0',
    ))
    result = version.find_latest_version(versions)
    assert str(result) == '3.0.1'


def test_ignore_prerelease_versions():
    versions = (semver.VersionInfo.parse(v) for v in (
            '0.0.10',
            '3.0.1',
            '1.0.0',
            '3.1.0-foo-bar',
    ))
    result = version.find_latest_version(versions, ignore_prerelease_versions=True)
    assert str(result) == '3.0.1'


def test_argument_validation():
    with pytest.raises(ValueError):
        version.process_version(version_str='invalid', operation='noop')

    # prerelease arg missing
    with pytest.raises(ValueError):
        version.process_version(version_str='1.2.3', operation='set_prerelease')

    # build_metadata_missing
    with pytest.raises(ValueError):
        version.process_version(version_str='1.2.3', operation='set_build_metadata')

    # metadata-len-smaller-than-zero
    with pytest.raises(ValueError):
        version.process_version(
        version_str='3.5.4',
        operation='set_build_metadata',
        build_metadata='someRandomString',
        build_metadata_length=-1
    )

    # prerelease present (in version) if using append_prerelease
    with pytest.raises(ValueError):
        version.process_version(
        version_str='3.5.4-foo',
        operation='append_prerelease'
    )

    # prerelease missing if using append_prerelease
    with pytest.raises(ValueError):
        version.process_version(
        version_str='3.5.4',
        operation='append_prerelease',
        prerelease='foo'
    )

    # version_str missing if using "set_verbatim"
    with pytest.raises(ValueError):
        version.process_version(
            version_str='3.1.4-foo',
            operation='set_verbatim',
            prerelease='baz'
        )


def test_noop():
    parsed = version.process_version(version_str='1.2.3-abc', operation='noop')
    assert parsed == '1.2.3-abc'


def test_set_build_metadata_length():
    parsed = version.process_version(
        version_str='1.3.5',
        operation='set_build_metadata',
        build_metadata='someRandomString',
        build_metadata_length=10
    )
    assert len(parsed.removeprefix('1.3.5+')) == 10


def test_set_prerelease_without_suffix():
    parsed = version.process_version(
        version_str='1.2.3',
        operation='set_prerelease',
        prerelease='dev'
    )
    assert parsed == '1.2.3-dev'


def test_set_build_metadata_without_suffix():
    parsed = version.process_version(
        version_str='3.3.3',
        operation='set_build_metadata',
        build_metadata='build'
    )
    assert parsed == '3.3.3+build'


def test_set_prerelease_with_prerelease():
    parsed = version.process_version(
        version_str='1.2.3-foo',
        operation='set_prerelease',
        prerelease='dev'
    )
    assert parsed == '1.2.3-dev'


def test_set_build_metadata_with_prerelease():
    parsed = version.process_version(
        version_str='3.3.3-foo',
        operation='set_build_metadata',
        build_metadata='build'
    )
    assert parsed == '3.3.3+build'


def test_set_prerelease_with_build_metadata():
    parsed = version.process_version(
        version_str='1.2.3+foo',
        operation='set_prerelease',
        prerelease='dev'
    )
    assert parsed == '1.2.3-dev'


def test_set_build_metadata_with_build_metadata():
    parsed = version.process_version(
        version_str='3.3.3+foo',
        operation='set_build_metadata',
        build_metadata='build'
    )
    assert parsed == '3.3.3+build'


def test_append_prerelease():
    parsed = version.process_version(
        version_str='4.9.16-foo',
        operation='append_prerelease',
        prerelease='bar',
    )
    assert parsed == '4.9.16-foo-bar'


def test_set_verbatim_with_verbatim_version():
    parsed = version.process_version(
        version_str='3.1.4-foo+bar',
        operation='set_verbatim',
        verbatim_version='master',
    )
    assert parsed == 'master'


def test_skip_patchlevel_zero():
    assert version.process_version(
        version_str='1.2.0',
        operation='noop',
        skip_patchlevel_zero=True,
    ) == '1.2.1'

    assert version.process_version(
        version_str='1.2.0',
        operation='noop',
        skip_patchlevel_zero=False,
    ) == '1.2.0'


def test_bumping():
    # major
    parsed = version.process_version(version_str='2.4.6', operation='bump_major')
    assert parsed == '3.0.0'

    # minor
    parsed = version.process_version(version_str='2.4.6', operation='bump_minor')
    assert parsed == '2.5.0'

    # patch
    parsed = version.process_version(version_str='2.4.6', operation='bump_patch')
    assert parsed  == '2.4.7'


def test_smallest_versions():
    # no filtering if keep >= amount of versions
    assert set(version.smallest_versions({'1.2.3', '2.3.4'}, keep=2)) == set()
    assert set(version.smallest_versions({'1.2.3', '2.3.4'}, keep=10)) == set()

    # keep greatest (returned versions are intended to be removed)
    assert set(version.smallest_versions({'1.2.3', '2.3.4', '3.0'}, keep=1)) == {'1.2.3', '2.3.4'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
import functools
import logging
import os
import re

from github3.exceptions import NotFoundError
import yaml

from ci.util import (
    LintingError,
    fail,
    lint_yaml,
    load_yaml,
    merge_dicts,
    not_empty,
    not_none,
    parse_yaml_file,
    urlparse,
)
from concourse.factory import RawPipelineDefinitionDescriptor
from model.base import ModelBase, NamedModelElement
import ccc.github
import ci.log
import concourse.paths
import model.concourse
import model.secret

ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


class JobMappingNotFoundError(ValueError):
    ''' Signals that no JobMapping was found for a GitHub repository when processing.
    '''
    pass


class DefinitionDescriptorPreprocessor:
    def process_definition_descriptor(self, descriptor):
        self._add_branch_to_pipeline_name(descriptor)
        return self._inject_main_repo(descriptor)

    def _add_branch_to_pipeline_name(self, descriptor):
        pipeline_name = descriptor.pipeline_name
        branch = descriptor.main_repo.get('branch').replace('/', '_')

        descriptor.pipeline_name = f'{pipeline_name}-{branch}'
        return descriptor

    def _inject_main_repo(self, descriptor):
        descriptor.override_definitions.append({
            'base_definition':
            {
                'repo': descriptor.main_repo
            }
        })

        return descriptor


class DefinitionEnumerator:
    def enumerate_definition_descriptors(self):
        raise NotImplementedError('subclasses must override')

    def _wrap_into_descriptors(
        self,
        repo_path,
        repo_hostname,
        branch,
        raw_definitions,
        secret_cfg,
        override_definitions={},
        job_mapping: model.concourse.JobMapping=None,
        target_team: str=None,
        pipeline_definition_committish: str=None,
    ) -> 'DefinitionDescriptor':
        if not target_team:
            target_team = self.job_mapping.team_name() # noqa

        for name, definition in raw_definitions.items():
            if not definition.get('render_pipeline', True):
                continue

            pipeline_definition = deepcopy(definition)
            yield DefinitionDescriptor(
                pipeline_name=name,
                pipeline_definition=pipeline_definition or {},
                main_repo={'path': repo_path, 'branch': branch, 'hostname': repo_hostname},
                concourse_target_cfg=self.cfg_set.concourse(), # noqa
                concourse_target_team=target_team,
                override_definitions=[override_definitions.get(name,{}),],
                secret_cfg=secret_cfg,
                job_mapping=job_mapping,
                pipeline_definition_committish=pipeline_definition_committish,
            )


class SimpleFileDefinitionEnumerator(DefinitionEnumerator):
    def __init__(
        self,
        definition_file,
        cfg_set,
        repo_path,
        repo_branch,
        repo_host='github.com',
        job_mapping: model.concourse.JobMapping=None,
        secret_cfg: model.secret.Secret=None,
    ):
        self.definition_file = definition_file
        self.repo_path = repo_path
        self.repo_branch = repo_branch
        self.repo_host = repo_host
        self.cfg_set = cfg_set
        self.job_mapping = job_mapping or model.concourse.JobMapping(
            name='dummy',
            raw_dict={'concourse_target_team': 'dummy'},
        )
        self.secret_cfg = secret_cfg or model.secret.Secret(
            name='dummy',
            raw_dict={
                'key': 'dummy',
                'cipher_algorithm': 'PLAINTEXT'
            },
        )

    def enumerate_definition_descriptors(self):
        logger.info('enumerating explicitly specified definition file')

        try:
            definitions = parse_yaml_file(self.definition_file)
            yield from self._wrap_into_descriptors(
                repo_path=self.repo_path,
                repo_hostname=self.repo_host,
                branch=self.repo_branch,
                raw_definitions=definitions,
                job_mapping=self.job_mapping,
                secret_cfg=self.secret_cfg,
            )
        except BaseException as e:
            yield DefinitionDescriptor(
                pipeline_name='<invalid YAML>',
                pipeline_definition={},
                main_repo={
                    'path': self.repo_path,
                    'branch': self.repo_branch,
                    'hostname': self.repo_host,
                },
                concourse_target_cfg=self.cfg_set.concourse(),
                concourse_target_team=self.job_mapping.team_name(),
                override_definitions=(),
                exception=e,
                secret_cfg=self.secret_cfg,
            )


class BranchCfg(ModelBase):
    def _required_attributes(self):
        return {'cfgs'}

    def cfg_entries(self):
        return (
            BranchCfgEntry(name=name, raw_dict=raw_dict)
            for name, raw_dict in self.raw.get('cfgs').items()
        )

    def cfg_entry_for_branch(self, branch):
        matching_entries = [
            entry for entry in self.cfg_entries() if entry.branch_matches(branch)
        ]
        if not matching_entries:
            return None

        # merge entries
        effective_branch_cfg = {}

        # order by optional attribute 'index' (random order if omitted)
        def entry_with_idx(idx, entry):
            if entry.index():
                idx = int(entry.index())
            return (idx, entry)

        indexed_entries = [entry_with_idx(idx, entry) for idx, entry in enumerate(matching_entries)]

        for _, entry in sorted(indexed_entries, key=lambda i: i[0]):
            effective_branch_cfg = merge_dicts(effective_branch_cfg, entry.raw)

        return BranchCfgEntry(name='merged', raw_dict=effective_branch_cfg)


class BranchCfgEntry(NamedModelElement):
    def _optional_attributes(self):
        return {'index'}

    def index(self):
        return self.raw.get('index')

    def branches(self):
        return self.raw.get('branches')

    def branch_matches(self, branch):
        for b in self.branches():
            if re.fullmatch(b, branch):
                return True
        return False

    def override_definitions(self):
        return self.raw.get('inherit', {})

    def __repr__(self):
        return f'BranchCfgEntry: {self.raw}'


class GithubDefinitionEnumeratorBase(DefinitionEnumerator):
    def _branch_cfg_or_none(
        self,
        repository,
    ):
        try:
            branch_cfg = repository.file_contents(
                path='branch.cfg',
                ref='refs/meta/ci',
            ).decoded.decode('utf-8')
            lint_yaml(branch_cfg)
        except NotFoundError:
            return None # no branch cfg present

        return BranchCfg(raw_dict=load_yaml(branch_cfg))

    def _determine_repository_branches(
        self,
        repository,
        branch: str = None,
    ):
        try:
            branch_cfg = self._branch_cfg_or_none(repository=repository)
        except LintingError as e:
            # some linting errors (and possibly warnings) present. Print warning and continue
            logger.warning(e)
            return

        if not branch_cfg:
            try:
                branch = branch or repository.default_branch
            except Exception:
                branch = 'master'
            yield branch, None
        else:
            if branch:
                cfg_entry = branch_cfg.cfg_entry_for_branch(branch)
                if cfg_entry:
                    # if the branch config has an entry that applies, use it
                    yield branch, cfg_entry
                else:
                    # since we explicitly requested the branch, use only pipeline-def here. This
                    # differs from the case when no branch is requested (where we would skip such
                    # branches)
                    yield branch, None
            else:
                yielded = False
                for branch in repository.branches():
                    cfg_entry = branch_cfg.cfg_entry_for_branch(branch.name)
                    if cfg_entry:
                        yielded = True
                        yield (branch.name, cfg_entry)
                if not yielded:
                    logger.warning(
                        'Applying branch.cfg resulted in no pipeline-definition for repository '
                        f"'{repository}'. Do the branches configured in the branch config exist?"
                    )

    def _scan_repository_for_definitions(
        self,
        repository,
        github_cfg,
        org_name,
        job_mapping: model.concourse.JobMapping=None,
        target_team: str=None,
        secret_cfg=None,
        branch: str=None,
    ) -> RawPipelineDefinitionDescriptor:

        repo_hostname = urlparse(github_cfg.http_url()).hostname
        repo_path = f'{org_name}/{repository.name}'

        if not target_team:
            target_team = self.job_mapping.team_name() # noqa

        try:
            branches_and_cfg_entries = [
                i for i in self._determine_repository_branches(
                    repository=repository,
                    branch=branch,
                )
            ]
        except (yaml.scanner.ScannerError, yaml.parser.ParserError) as e:
            yield DefinitionDescriptor(
                pipeline_name='<invalid YAML>',
                pipeline_definition=None,
                main_repo={'path': repo_path, 'branch': 'refs/meta/ci', 'hostname': repo_hostname},
                concourse_target_cfg=self.cfg_set.concourse(), # noqa
                concourse_target_team=target_team,
                override_definitions=(),
                exception=e,
                secret_cfg=secret_cfg,
            )
            return # nothing else to yield in case parsing the branch cfg failed

        for branch_name, cfg_entry in branches_and_cfg_entries:
            try:
                definitions = repository.file_contents(
                    path='.ci/pipeline_definitions',
                    ref=branch_name
                )
                pipeline_definition_committish = repository.ref(
                    ref=f'heads/{branch_name}'
                ).object.sha
            except NotFoundError:
                continue # no pipeline definition for this branch

            override_definitions = cfg_entry.override_definitions() if cfg_entry else {}

            try:
                decoded_definitions = definitions.decoded.decode('utf-8')
                lint_yaml(decoded_definitions)
                definitions = load_yaml(decoded_definitions)

            except BaseException as e:
                yield DefinitionDescriptor(
                    pipeline_name='<invalid YAML>',
                    pipeline_definition={},
                    main_repo={'path': repo_path, 'branch': branch_name, 'hostname': repo_hostname},
                    concourse_target_cfg=self.cfg_set.concourse(), # noqa
                    concourse_target_team=target_team,
                    override_definitions=(),
                    exception=e,
                    secret_cfg=secret_cfg,
                )
                return # nothing else to yield in case parsing failed

            # handle inheritance
            definitions = merge_dicts(definitions, override_definitions)

            # hacky: only set from GithubRepositoryDefinitionEnumerator
            target_team = getattr(self, '_target_team', None)

            yield from self._wrap_into_descriptors(
                repo_path='/'.join([org_name, repository.name]),
                repo_hostname=repo_hostname,
                branch=branch_name,
                raw_definitions=definitions,
                override_definitions=override_definitions,
                target_team=target_team,
                secret_cfg=secret_cfg,
                job_mapping=job_mapping,
                pipeline_definition_committish=pipeline_definition_committish,
            )


class GithubRepositoryDefinitionEnumerator(GithubDefinitionEnumeratorBase):
    def __init__(
        self,
        repository_url:str,
        cfg_set,
        target_team: str=None,
        job_mapping=None,
        branch: str=None,
    ):
        self._repository_url = ci.util.urlparse(not_none(repository_url))
        self._repo_host = self._repository_url.hostname
        self.cfg_set = not_none(cfg_set)
        self._target_team = target_team
        concourse_cfg = cfg_set.concourse()
        job_mapping_set = cfg_set.job_mapping(concourse_cfg.job_mapping_cfg_name())
        self.job_mapping = job_mapping
        self.branch = branch

        org_name, repo_name = self._repository_url.path.lstrip('/').split('/')

        if target_team:
            return # if tgt-team is explicitly configured, we do not need to look it up

        if self.job_mapping:
            return

        for job_mapping in job_mapping_set.job_mappings().values():
            for org in job_mapping.github_organisations():
                if org.org_name() != org_name:
                    continue
                if not org.repository_matches(repo_name):
                    continue
                github_cfg = cfg_set.github(org.github_cfg_name())
                if github_cfg.matches_hostname(self._repo_host):
                    self.job_mapping = job_mapping
                    return
        else:
            raise JobMappingNotFoundError(
                f"could not find matching job-mapping for repo '{repo_name}' in org '{org_name}'"
            )

    def enumerate_definition_descriptors(self):
        github_cfg = ccc.github.github_cfg_for_repo_url(
            repo_url=self._repository_url,
        )
        github_api = ccc.github.github_api(github_cfg=github_cfg)
        github_org, github_repo = self._repository_url.path.lstrip('/').split('/')
        repository = github_api.repository(github_org, github_repo)

        if self.job_mapping and self.job_mapping.secret_cfg():
            secret_cfg = self.cfg_set.secret(self.job_mapping.secret_cfg())
        else:
            secret_cfg = None

        yield from self._scan_repository_for_definitions(
            repository=repository,
            github_cfg=github_cfg,
            org_name=github_org,
            target_team=self._target_team,
            secret_cfg=secret_cfg,
            job_mapping=self.job_mapping,
            branch=self.branch,
        )


class GithubOrganisationDefinitionEnumerator(GithubDefinitionEnumeratorBase):
    def __init__(
        self,
        job_mapping,
        cfg_set,
        repository_filter: callable=None,
    ):
        self.job_mapping = not_none(job_mapping)
        self.cfg_set = not_none(cfg_set)
        self.repository_filter = repository_filter

    def enumerate_definition_descriptors(self):
        executor = ThreadPoolExecutor(max_workers=16)

        # scan github repositories
        for github_org_cfg in self.job_mapping.github_organisations():
            github_cfg = self.cfg_set.github(github_org_cfg.github_cfg_name())
            github_org_name = github_org_cfg.org_name()
            logger.info(
                f'scanning github organisation {github_org_name}'
                f' with github_cfg {github_cfg.name()}'
            )

            github_api = ccc.github.github_api(github_cfg)
            github_org = github_api.organization(github_org_name)

            if self.job_mapping.secret_cfg():
                secret_cfg = self.cfg_set.secret(self.job_mapping.secret_cfg())
            else:
                secret_cfg = None

            scan_repository_for_definitions = functools.partial(
                self._scan_repository_for_definitions,
                github_cfg=github_cfg,
                org_name=github_org_name,
                secret_cfg=secret_cfg,
                job_mapping=self.job_mapping,
            )

            matching_repositories = (
                repo for repo in github_org.repositories()
                if github_org_cfg.repository_matches(repo.name)
                   and (not self.repository_filter or self.repository_filter(repo))
            )

            for definition_descriptors in executor.map(
                scan_repository_for_definitions,
                matching_repositories,
            ):
                yield from definition_descriptors


class DefinitionDescriptor:
    '''
    Container type holding the result of a pipeline rendering and additional
    metadata required to deploy as concrete Concourse Pipeline.
    '''
    def __init__(
        self,
        pipeline_name,
        pipeline_definition,
        main_repo,
        concourse_target_cfg: model.concourse.ConcourseConfig,
        concourse_target_team,
        secret_cfg,
        job_mapping: model.concourse.JobMapping=None,
        override_definitions=[{},],
        exception=None,
        pipeline_definition_committish: str=None,
    ):
        try:
            self.pipeline_name = not_empty(pipeline_name)
            self.pipeline_definition = not_none(pipeline_definition)
            self.main_repo = not_none(main_repo)
            self.concourse_target_cfg = not_none(concourse_target_cfg)
            self.concourse_target_team = not_none(concourse_target_team)
            self.override_definitions = not_none(override_definitions)
            self.secret_cfg = secret_cfg
            self.job_mapping = job_mapping
            self.pipeline_definition_committish = pipeline_definition_committish
        except Exception as e:
            raise ValueError(
                f'{e=} missing value: {pipeline_name=} {pipeline_definition=} {main_repo=} '
                f'{concourse_target_cfg=} {concourse_target_team=} {override_definitions=}'
            )
        self.exception = exception

    def template_name(self):
        return self.pipeline_definition.get('template', 'default')

    def effective_pipeline_name(self):
        return '-'.join((self.pipeline_name, self.main_repo['branch']))

    def concourse_target(self):
        return (self.concourse_target_cfg, self.concourse_target_team)

    def concourse_target_key(self):
        return '{n}:{t}'.format(
            n=self.concourse_target_cfg.name(),
            t=self.concourse_target_team
        )

    def __repr__(self):
        return f'{self.__class__}: {self.pipeline_name}'


class TemplateRetriever:
    '''
    Provides mako templates by name. Templates are cached.
    '''

    def __init__(
        self,
        template_path=concourse.paths.template_dir
    ):
        if type(template_path) == str:
            self.template_path = (template_path,)
        else:
            self.template_path = template_path

    @functools.lru_cache()
    def template_file(self, template_name):
        # TODO: do not hard-code file name extension
        template_file_name = template_name + '.mako'
        for path in self.template_path:
            for dirpath, _, filenames in os.walk(path):
                if template_file_name in filenames:
                    return os.path.join(dirpath, template_file_name)
        fail(
            'could not find template {t}, tried in {p}'.format(
                t=str(template_name),
                p=','.join(map(str, self.template_path))
            )
        )

    @functools.lru_cache()
    def template_contents(self, template_name='default'):
        with open(self.template_file(template_name=template_name)) as f:
            return f.read()



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    ModelBase,
    NamedModelElement,
)


class DockerImageConfig(ModelBase):
    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'image_name',
            'image_tag',
        ]

    def image_name(self):
        return self.raw['image_name']

    def image_tag(self):
        return self.raw['image_tag']

    def image_reference(self):
        return f'{self.image_name()}:{self.image_tag()}'


class MitmLoggingConfig(ModelBase):
    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'els_config',
            'els_index',
        ]

    def els_config_name(self):
        return self.raw['els_config']

    def els_index_name(self):
        return self.raw['els_index']


class ProxyConfig(NamedModelElement):
    '''Encompasses all configuration necessary for the deployment of a MitM-Proxy alongside
    Concourse.

    Consists of two sub-configs: the config for the init container required to setup our Pod for the
    mitm-proxy and the configuration of the MitM-proxy.
    '''
    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'mitm_proxy',
            'sidecar_image',
            'job_mapping_name',
        ]

    def mitm_proxy(self):
        return MitmProxyConfig(raw_dict=self.raw['mitm_proxy'])

    def sidecar_image(self):
        return DockerImageConfig(raw_dict=self.raw['sidecar_image'])

    def job_mapping_name(self) -> str:
        return self.raw.get('job_mapping_name')


class MitmProxyConfig(DockerImageConfig):
    '''A combination of docker image reference and MitM-Config.

    The content of the config attribute is a mapping of MitM-Proxy options to their values. An
    annotated example of all options can be obtained by running 'mitmproxy --options'.
    '''
    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'config',
            'logging',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'filter_config',
        ]

    def config(self):
        return self.raw['config']

    def logging(self):
        return MitmLoggingConfig(self.raw['logging'])

    def filter_config(self):
        config = self.raw.get('filter_config')
        if not config:
            return None
        return MitmFilterConfig(config)


class MitmFilterConfig(ModelBase):
    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'whitelisted_hosts',
            'blacklisted_hosts',
        ]

    def whitelisted_hosts(self):
        return self.raw.get('whitelisted_hosts', ())

    def blacklisted_hosts(self):
        return self.raw.get('blacklisted_hosts', ())



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
    ModelBase,
)


class CCEEProject(ModelBase):
    def region(self):
        return self.raw.get('region')

    def name(self):
        return self.raw.get('name')

    def domain(self):
        return self.raw.get('domain')

    def auth_url(self):
        return self.raw.get('auth_url')


class CCEEConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def _required_attributes(self):
        return ['credentials']

    def credentials(self):
        return BasicCredentials(self.raw['credentials'])

    def projects(self):
        return [
            CCEEProject(raw_dict=project_dict) for project_dict in self.raw['projects']
        ]

    def _defaults_dict(self):
        return {
            'projects': (),
        }

    def _optional_attributes(self):
        return (
            'projects',
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import collections
import datetime
import enum
import functools
import logging
import os
import pathlib
import shutil
import string
import sys
import typing
import urllib.parse
import yaml

try:
    import yamllint
    import yamllint.config
    _have_yamllint = True
except ImportError:
    _have_yamllint = False

import deprecated
try:
    import termcolor
    _have_termcolor = True
except ImportError:
    _have_termcolor = False

import ci.paths


logger = logging.getLogger(__name__)


class Failure(RuntimeError, ValueError):
    pass


class LintingError(Failure):
    pass


class LintingResult:
    def __init__(self, problems):
        if not _have_yamllint:
            raise RuntimeError('need to install `yamllint` in order to use')

        problems_dict = collections.defaultdict(list)
        for p in problems:
            problems_dict[self._normalise_problem_level(p)].append(p)
        self.problems_dict = problems_dict

    def _normalise_problem_level(self, p):
        # Yamllint uses a dict with mappings for both int->level and level->int. Consistently use
        # the int representation
        if isinstance(p.level, str):
            return yamllint.linter.PROBLEM_LEVELS[p.level]
        else:
            return p.level

    def problems(self):
        return self.problems_dict.items()

    def has_problems(self):
        return bool(self.problems_dict)

    def max_level(self):
        if self.problems:
            max(self.problems_dict.keys())
        return 0

    def __str__(self):
        if not self.has_problems():
            return "No linting problems found"

        return '\n'.join(s for s in [
            f'[{yamllint.linter.PROBLEM_LEVELS[l]}] {p}'
            for l in self.problems_dict.keys()
            for p in self.problems_dict[l]
        ])


def _set_cli(is_cli: bool):
    ctx().args._cli = is_cli
    global Failure
    if is_cli:
        class Failure(SystemExit):
            pass
    else:
        class Failure(RuntimeError):
            pass


def existing_file(path):
    if isinstance(path, pathlib.Path):
        is_file = path.is_file()
    else:
        is_file = os.path.isfile(path)
    if not is_file:
        fail('not an existing file: ' + str(path))
    return path


def existing_dir(path: str):
    if isinstance(path, pathlib.Path):
        is_dir = path.is_dir()
    else:
        is_dir = os.path.isdir(path)
    if not is_dir:
        fail('not an existing directory: ' + str(path))
    return path


def check_type(instance, type):
    if not isinstance(instance, type):
        fail('{i} is not an instance of {t}'.format(i=instance, t=type))
    return instance


def gardener_cicd_libs_version():
    versionfile_file_path = ci.paths.version_file
    with open(versionfile_file_path, 'rt') as f:
        return f.readline()


class CliHint:
    def __init__(self, typehint=str, *args, **kwargs):
        self.argparse_args = dict(*args, **kwargs)
        self.typehint = typehint


class CliHints:
    '''
    predefined cli hint instances
    '''
    @staticmethod
    def existing_file(help: str='an existing file', **kwargs):
        return CliHint(type=existing_file, help=help, **kwargs)

    @staticmethod
    def yaml_file(help: str='an existing YAML file', **kwargs):
        return CliHint(type=parse_yaml_file, help=help, **kwargs)

    @staticmethod
    def existing_dir(help: str='an existing directory', **kwargs):
        return CliHint(type=existing_dir, help=help, **kwargs)

    @staticmethod
    def non_empty_string(help:str = 'a non-empty string', **kwargs):
        return CliHint(type=not_empty, help=help, **kwargs)


def ctx():
    # late import because the ctx module is altered after all existing modules have
    # already been imported
    import ctx
    return ctx


def _quiet():
    return ctx().args and ctx().args.quiet


def _verbose():
    return ctx().args and ctx().args.verbose


def _print(msg, colour, outfh=sys.stdout):
    if not msg:
        return
    if not outfh.isatty() or not _have_termcolor:
        outfh.write(msg + '\n')
    else:
        outfh.write(termcolor.colored(msg, colour) + '\n')

    outfh.flush()


@deprecated.deprecated
def error(msg=None):
    if _quiet():
        return
    if msg:
        _print('ERROR: ' + str(msg), colour='red', outfh=sys.stderr)


@deprecated.deprecated
def fail(msg=None):
    if msg:
        _print('ERROR: ' + str(msg), colour='red', outfh=sys.stderr)
    raise Failure(1)


@deprecated.deprecated
def info(msg:str):
    if _quiet():
        return
    if msg:
        _print('INFO: ' + str(msg), colour='cyan', outfh=sys.stdout)


@deprecated.deprecated
def warning(msg:str):
    if _quiet():
        return
    if msg:
        _print('WARNING: ' + str(msg), colour='yellow', outfh=sys.stderr)


@deprecated.deprecated
def verbose(msg:str):
    if not _verbose():
        return
    if msg:
        _print('VERBOSE: ' + msg, colour=None, outfh=sys.stdout)


@deprecated.deprecated
def success(msg:str):
    if msg:
        _print('SUCCESS: ' + msg, colour='green', outfh=sys.stdout)


def not_empty(value):
    if not value or len(value) == 0:
        fail('passed value must not be empty')
    return value


def not_none(value):
    if value is None:
        fail('passed value must not be None')
    return value


def load_yaml(stream, lint=False, linter_config=None):
    '''
    Parses YAML from the given stream in a (by default) safe manner.

    In addition to using SafeLoader, a mitigation against YAML Bombs (Billion Laughs Attack) is
    applied (by limiting amount of allowed elements)

    @raises ValueError if YAML Bomb was (heuristically) detected.
    '''
    if lint:
        if not _have_yamllint:
            raise RuntimeError('need to install `yamllint` in order to use')

        # redefine stream, as yamllint will read() this, resulting in stream being empty
        # when parsing later
        stream = stream.read()
        linter_config = yamllint.config.YamlLintConfig(yaml.dump(not_none(linter_config)))
        linting_result = LintingResult(yamllint.linter.run(input=stream, conf=linter_config))
        _print_linting_findings(linting_result)

    parsed = yaml.safe_load(stream)
    _count_elements(parsed)
    return parsed


def parse_yaml_file(
    path: str,
    lint: bool=False,
) -> dict:
    '''
    parses yaml file from local file system to dict
    '''
    if lint:
        lint_yaml_file(path)

    with open(path) as f:
        parsed = yaml.load(f, Loader=yaml.SafeLoader)
        # mitigate yaml bomb
        _count_elements(parsed)
        return parsed


def _count_elements(value, count=0, max_elements_count=100000):
    '''
    recursively counts elements contained in the given value. Before each recursion step,
    the amount of encountered elements is checked against a maximum allowed elements count.
    If said threshold is exceeded, recursion is aborted and a `ValueError` is raised.

    This function is intended to be used as a mitigation against "Billion laughs attack"
    (https://en.wikipedia.org/wiki/Billion_laughs_attack).

    @param value: typically a dict or a list. Other types will yield a count of 1
    '''
    if count > max_elements_count:
        raise ValueError('dict too large')

    if not isinstance(value, dict):
        if isinstance(value, list):
            leng = 0
            for e in value:
                leng += _count_elements(e, count=count+leng)
            return leng
        else:
            return 1

    leng = 0

    for value in value.values():
        leng += _count_elements(value, count=count+leng)

    return leng


def lint_yaml_file(path, linter_config: dict={'extends': 'relaxed'}):
    if not _have_yamllint:
        raise RuntimeError('need to install yamllint in order to use')

    existing_file(path)
    logger.info(f'linting YAML file: {path}')

    with open(path) as f:
        linting_result = _lint_yaml(f.read(), linter_config)

    if not linting_result.has_problems():
        return

    _print_linting_findings(linting_result)

    if linting_result.max_level() >= yamllint.linter.PROBLEM_LEVELS['error']:
        raise LintingError('Found some Errors while linting. See above.')


def _print_linting_findings(linting_result: LintingResult):
    if not _have_yamllint:
        raise RuntimeError('need to install yamllint in order to use')
    for level, problems in linting_result.problems():
        if level < yamllint.linter.PROBLEM_LEVELS['error']:
            for p in problems:
                warning(p)
        else:
            for p in problems:
                error(p)


def _lint_yaml(input, config):
    if not _have_yamllint:
        raise RuntimeError('need to install yamllint in order to use')
    cfg = yamllint.config.YamlLintConfig(yaml.dump(config))
    return LintingResult(yamllint.linter.run(input=input, conf=cfg))


def lint_yaml(input, config={'extends': 'relaxed'}):
    if not _have_yamllint:
        raise RuntimeError('need to install yamllint in order to use')

    linting_result = _lint_yaml(input=input, config=config)

    if not linting_result.has_problems():
        return

    if linting_result.max_level() > yamllint.linter.PROBLEM_LEVELS['warning']:
        raise LintingError(f'linter found errors {linting_result=}')


def random_str(
    prefix: typing.Optional[str]=None,
    length: int=12,
    alphabet: typing.Optional[str]=string.ascii_lowercase,
) -> str:
    import random
    if prefix:
        length -= len(prefix)
    else:
        prefix = ''
    return prefix + ''.join(random.choice(alphabet) for _ in range(length)) # nosec B311


def check_env(name: str):
    '''
    returns: the specified environment variable's value.
    raises: ci.util.Failure if no environment variable with the given name is defined
    '''
    not_none(name)
    if name in os.environ:
        return os.environ[name]
    fail('env var {n} must be set'.format(n=name))


def _running_on_ci():
    '''
    heuristically determines whether or not the caller is running inside a central
    CI/CD job.
    '''
    return 'CC_ROOT_DIR' in os.environ


def current_config_set_name():
    if not _running_on_ci():
        raise RuntimeError('must only be called within CI/CD')
    return check_env('CONCOURSE_CURRENT_CFG')


def _root_dir():
    if not _running_on_ci():
        raise RuntimeError('must only be called within CI/CD')
    return check_env('CC_ROOT_DIR')


def urljoin(*parts):
    if len(parts) == 1:
        return parts[0]
    first = parts[0]
    last = parts[-1]
    middle = parts[1:-1]

    first = first.rstrip('/')
    middle = list(map(lambda s: s.strip('/'), middle))
    last = last.lstrip('/')

    return '/'.join([first] + middle + [last])


def urlparse(url: str) -> urllib.parse.ParseResult:
    if not '://' in url:
        url = f'x://{url}'

    return urllib.parse.urlparse(url)


def file_extension_join(path: str, extension: str) -> str:
    return '.'.join((path, extension))


def which(cmd_name: str) -> str:
    '''
    wrapper around shutil.which that calls ci.util.fail if the requested executable is not
    found in the PATH.
    '''
    cmd_path = shutil.which(cmd_name)
    if not cmd_path:
        fail("{cmd} not found in PATH".format(cmd=cmd_name))
    return cmd_path


def merge_dicts(base: dict, *other: dict, list_semantics='merge'):
    '''
    merges copies of the given dict instances and returns the merge result.
    The arguments remain unmodified. However, it must be possible to copy them
    using `copy.deepcopy`.

    Merging is done using the `deepmerge` module. In case of merge conflicts, values from
    `other` overwrite values from `base`.

    By default, different from the original implementation, a merge will be applied to
    lists. This results in deduplication retaining element order. The elements from `other` are
    appended to those from `base`.
    '''

    not_none(base)
    not_empty(other)

    from deepmerge import Merger

    if list_semantics == 'merge':
        # monkey-patch merge-strategy for lists
        list_merge_strategy = Merger.PROVIDED_TYPE_STRATEGIES[list]
        list_merge_strategy.strategy_merge = lambda c, p, base, other: \
            list(base) + [e for e in other if e not in base]

        strategy_cfg = [(list, ['merge']), (dict, ['merge'])]
        merger = Merger(strategy_cfg, ['override'], ['override'])
    elif list_semantics is None:
        strategy_cfg = [(dict, ['merge'])]
        merger = Merger(strategy_cfg, ['override'], ['override'])
    else:
        raise NotImplementedError

    from copy import deepcopy

    return functools.reduce(
        lambda b, o: merger.merge(b, deepcopy(o)),
        [base, *other],
        {},
    )


class FluentIterable:
    ''' a fluent object stream processing chain builder inspired by guava's FluentIterable

    Example:
        result = FluentIterable(items=(1,2,3))
            .filter(lambda e: e < 2)
            .map(lambda e: e * 2)
            .as_generator()

    '''

    def __init__(self, items):
        def starter():
            yield from items
        self.ops = [starter]

    def filter(self, filter_func):
        last_op = self.ops[-1]

        def f():
            yield from filter(filter_func, last_op())

        self.ops.append(f)
        return self

    def map(self, map_func):
        last_op = self.ops[-1]

        def m():
            yield from map(map_func, last_op())

        self.ops.append(m)
        return self

    def as_generator(self):
        return self.ops[-1]()

    def as_list(self):
        return list(self.as_generator())


def dict_factory_enum_serialisiation(data):

    def convert_value(obj):
        if isinstance(obj, enum.Enum):
            return obj.value
        return obj

    return dict((k, convert_value(v)) for k, v in data)


def dict_to_json_factory(data):
    '''
    converts enums to their respective values and datetime.date[time] stamps
    to iso8601 timestamp strings
    '''

    def convert_value(obj):
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        elif isinstance(obj, enum.Enum):
            return obj.value
        return obj

    return dict((k, convert_value(v)) for k, v in data)


class MultilineYamlDumper(yaml.SafeDumper):
    def represent_data(self, data):
        # by default, the SafeDumper includes an extra empty line for each line in the data for
        # string-blocks. As all provided ways to configure the dumper differently affect all
        # rendered types we create our own Dumper.
        if isinstance(data, str) and '\n' in data:
            return self.represent_scalar(u'tag:yaml.org,2002:str', data, style='|')
        # Also, don't include keys with None/null values.
        if data is None:
            return self.represent_scalar('tag:yaml.org,2002:null', '')
        return super().represent_data(data)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dacite

import model.base


class AwsProfile(model.base.NamedModelElement):
    def region(self):
        return self.raw['region']

    def access_key_id(self):
        return self.raw['access_key_id']

    def secret_access_key(self):
        return self.raw['secret_access_key']

    def rotation_cfg(self) -> model.base.CfgElementReference:
        '''
        used to specify cfg-element to use for cross-rotation
        '''
        raw = self.raw.get('rotation_cfg')
        if raw:
            return dacite.from_dict(
                data_class=model.base.CfgElementReference,
                data=raw,
            )

        return None

    def iam_user_name(self):
        '''
        used to specify own iam user when rotating with cross-rotation
        '''
        return self.raw.get('iam_user_name')

    def _required_attributes(self):
        static_attributes = [
            'region',
            'access_key_id',
            'secret_access_key',
        ]
        dynamic_attributes = ['iam_user_name'] if self.rotation_cfg() else []

        return static_attributes + dynamic_attributes



'''
Release Notes Collector

Collects release notes for a given version from a ComponentDescriptor.

For a new patch release, all commits between the last patch tag up to the
current patch tag are included in the release notes. For a new minor release,
all commits between the last minor release (e.g. 1.63.0) up to the current
version are included and all commits before the last minor are removed.  It
doesn't matter on which branch the release tags are created, as long as they
share an ancestor.

This package is a re-implementation of the current release note collector
(https://github.com/gardener/cc-utils/tree/master/github/release_notes) to
avoid code freezes, which currently occur because in the current version you
have to release on a branch, while in the re-implementation you can release on
any branch.

See https://github.com/gardener/cc-utils/blob/master/doc/release_notes.rst for
more info.
'''



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import logging

import concourse.enumerator
import concourse.paths
import concourse.replicator
import model
import model.concourse
import model.webhook_dispatcher

logger = logging.getLogger(__name__)


def validate_repository_pipelines(
    repo_url: str,
    cfg_set: model.ConfigurationSet,
    whd_cfg: model.webhook_dispatcher.WebhookDispatcherConfig,
    branch: str=None,
    job_mapping=None,
):
    logger.info(f'validating pipeline(s) for {repo_url=} using {cfg_set.name()}')

    preprocessor = concourse.enumerator.DefinitionDescriptorPreprocessor()
    template_retriever = concourse.enumerator.TemplateRetriever()
    renderer = concourse.replicator.Renderer(
        template_retriever=template_retriever,
        template_include_dir=concourse.paths.template_include_dir,
        cfg_set=cfg_set,
        render_origin=concourse.replicator.RenderOrigin.WEBHOOK_DISPATCHER,
    )

    repo_enumerator = concourse.enumerator.GithubRepositoryDefinitionEnumerator(
        repository_url=repo_url,
        cfg_set=cfg_set,
        branch=branch,
        job_mapping=job_mapping,
    )
    deployer = concourse.replicator.NoOpDeployer()
    results_processor = concourse.replicator.PipelineValidationResultProcessor()

    replicator = concourse.replicator.PipelineReplicator(
        definition_enumerators=[repo_enumerator],
        descriptor_preprocessor=preprocessor,
        definition_renderer=renderer,
        definition_deployer=deployer,
        result_processor=results_processor,
    )

    logger.info('awaiting validation results')
    replicator.replicate()


def replicate_repository_pipelines(
    repo_url: str,
    cfg_set: model.ConfigurationSet,
    whd_cfg: model.webhook_dispatcher.WebhookDispatcherConfig,
):
    logger.info(f'replicating pipeline(s) for {repo_url=} using {cfg_set.name()}')

    preprocessor = concourse.enumerator.DefinitionDescriptorPreprocessor()
    template_retriever = concourse.enumerator.TemplateRetriever()
    renderer = concourse.replicator.Renderer(
        template_retriever=template_retriever,
        template_include_dir=concourse.paths.template_include_dir,
        cfg_set=cfg_set,
        render_origin=concourse.replicator.RenderOrigin.WEBHOOK_DISPATCHER,
    )

    repo_enumerator = concourse.enumerator.GithubRepositoryDefinitionEnumerator(
        repository_url=repo_url,
        cfg_set=cfg_set,
    )
    deployer = concourse.replicator.ConcourseDeployer(
        cfg_set=cfg_set,
        unpause_pipelines=False,
        expose_pipelines=True,
    )
    results_processor = concourse.replicator.ReplicationResultProcessor(
        cfg_set=cfg_set,
        unpause_new_pipelines=True,
        remove_pipelines=False,
        reorder_pipelines=False,
    )

    replicator = concourse.replicator.PipelineReplicator(
        definition_enumerators=[repo_enumerator],
        descriptor_preprocessor=preprocessor,
        definition_renderer=renderer,
        definition_deployer=deployer,
        result_processor=results_processor,
    )

    logger.info('awaiting replication-results')
    result = replicator.replicate()
    logger.info(f'{result=}')



import collections.abc
import dataclasses
import logging

import ocm

import ccc.oci
import cnudie.iter as ci
import oci.model

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ValidationError:
    node: ci.Node
    error: str

    @property
    def as_error_message(self):
        def _node_id(node: ci.Node | ocm.Component):
            if isinstance(node, ci.ComponentNode):
                return f'{node.component.name}:{node.component.version}'
            elif isinstance(node, ocm.Component):
                component = node
                return f'{component.name}:{component.version}'
            elif isinstance(node, ci.ResourceNode):
                artefact = node.resource
            elif isinstance(node, ci.SourceNode):
                artefact = node.source
            else:
                raise TypeError(node)

            return f'{artefact.name}:{artefact.version}'

        node_id_path = '/'.join((_node_id(node) for node in self.node.path))

        return f'{node_id_path}: {self.error}'


def validate_resource_node(
    node: ci.ResourceNode,
) -> ValidationError | None:
    resource = node.resource
    if resource.type != ocm.ArtefactType.OCI_IMAGE:
        return

    access: ocm.OciAccess = resource.access
    image_reference = access.imageReference

    try:
        image_reference = oci.model.OciImageReference.to_image_ref(image_reference)
        if not image_reference.has_tag:
            return ValidationError(
                node=node,
                error=f'Invalid ImageReference (missing tag): {image_reference}',
            )
    except ValueError:
        # cannot perform checks in image itself using invalid image-ref
        return ValidationError(
            node=node,
            error=f'Invalid ImageReference: {image_reference}',
        )

    oci_client = ccc.oci.oci_client()

    if not oci_client.head_manifest(
        image_reference=image_reference,
        absent_ok=True,
        accept=oci.model.MimeTypes.prefer_multiarch,
    ):
        return ValidationError(
            node=node,
            error=f'{image_reference=} does not exist',
        )


def iter_violations(
    nodes: collections.abc.Iterable[ci.Node],
) -> collections.abc.Generator[ValidationError, None, None]:
    for node in nodes:
        if isinstance(node, ci.ComponentNode):
            continue # no validation, yet
        elif isinstance(node, ci.SourceNode):
            continue # no validation, yet
        elif isinstance(node, ci.ResourceNode):
            if (validation_error := validate_resource_node(node=node)):
                yield validation_error
        else:
            raise ValueError(node)



import os
import stat
import tempfile
import unittest

from unittest.mock import MagicMock

import test_utils

from concourse.client.model import BuildStatus
from concourse.steps import step_def
from concourse.model.base import ScriptType
from concourse.model.job import JobVariant
from concourse.model.resources import (
    RepositoryConfig,
    ResourceRegistry,
    ResourceIdentifier,
    Resource,
)
from concourse.model.step import PipelineStep
from concourse.steps import notification
from concourse.model.traits.notifications import (
    NotificationCfgSet,
    NotificationTriggeringPolicy,
)


class NotificationStepTest(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.meta_dir = os.path.join(self.tmp_dir.name, 'meta')
        os.mkdir(self.meta_dir)
        test_utils.populate_meta_dir(self.meta_dir)
        self.on_error_dir = os.path.join(self.tmp_dir.name, 'on_error_dir')
        os.mkdir(self.on_error_dir)

        self.job_step = PipelineStep(
            name='step1',
            is_synthetic=False,
            script_type=ScriptType.BOURNE_SHELL,
            raw_dict={},
        )
        self.job_step._notifications_cfg = NotificationCfgSet('default', {}, type_name='cfg_set')
        resource_registry = ResourceRegistry()
        meta_resource_identifier = ResourceIdentifier(
            type_name='meta',
            base_name='a_job',
            branch_name='a_branch',
        )
        meta_resource = Resource(resource_identifier=meta_resource_identifier, raw_dict={})
        resource_registry.add_resource(meta_resource)
        self.job_variant = JobVariant(
            name='a_job',
            raw_dict={},
            resource_registry=resource_registry
        )

        # Set a main repository manually
        test_repo_logical_name = 'test-repository'
        self.job_variant._repos_dict = {}
        self.job_variant._repos_dict[test_repo_logical_name] = RepositoryConfig(
            raw_dict={
                'branch': 'master',
                'hostname': 'github.foo.bar',
                'path': 'test/repo'
            },
            logical_name=test_repo_logical_name,
            qualifier=None,
            is_main_repo=True
        )
        self.job_variant._main_repository_name = test_repo_logical_name

        self.job_variant._traits_dict = {}
        self.cfg_set = MagicMock()
        self.github_cfg = MagicMock()
        self.github_cfg.name = MagicMock(return_value='github_cfg')
        self.email_cfg = MagicMock()
        self.email_cfg.name = MagicMock(return_value='email_cfg')
        self.cfg_set.github = MagicMock(return_value=self.github_cfg)
        self.cfg_set.email = MagicMock(return_value=self.email_cfg)
        ctx_repo_mock = MagicMock(return_value='repo_url')
        ctx_repo_mock.base_url = MagicMock(return_value='repo_url')
        self.cfg_set.ctx_repository = MagicMock(return_value=ctx_repo_mock)

        self.render_step = step_def('notification')

        self.old_cwd = os.getcwd()
        os.chdir(self.tmp_dir.name)

    def tearDown(self):
        self.tmp_dir.cleanup()
        os.chdir(self.old_cwd)

    def test_render_and_compile(self):
        # as a smoke-test, just try to render
        step_snippet = self.render_step(
            job_step=self.job_step,
            job_variant=self.job_variant,
            cfg_set=self.cfg_set,
            repo_cfgs=(),
            subject='mail_subject1',
            indent=0
        )

        # try to compile (-> basic syntax check)
        return compile(step_snippet, 'notification', 'exec')


class NotificationStepLibTest(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.old_cwd = os.getcwd()
        os.chdir(self.tmp_dir.name)

    def tearDown(self):
        self.tmp_dir.cleanup()
        os.chdir(self.old_cwd)

    def test_job_url(self):
        v = {
            'atc-external-url': 'f://x',
            'build-team-name': 'team',
            'build-pipeline-name': 'pl',
            'build-job-name': 'bjn',
            'build-name': 'bn'
        }
        examinee = notification.job_url
        result = examinee(v)

        self.assertEqual(result, 'f://x/teams/team/pipelines/pl/jobs/bjn/builds/bn')

    def test_should_notify(self):
        examinee = notification.should_notify

        # mock away `determine_previous_build_status` (previous build "succeeded")
        build_status_mock = MagicMock(return_value=BuildStatus.SUCCEEDED)

        # test policies in case previous build succeeded
        assert examinee(
                NotificationTriggeringPolicy.ONLY_FIRST,
                meta_vars={},
                determine_previous_build_status=build_status_mock,
                cfg_set=None,
        )
        assert examinee(
                NotificationTriggeringPolicy.ALWAYS,
                meta_vars={},
                determine_previous_build_status=build_status_mock,
                cfg_set=None,
        )
        assert not examinee(
                NotificationTriggeringPolicy.NEVER,
                meta_vars={},
                determine_previous_build_status=build_status_mock,
                cfg_set=None,
        )

        # test policies in case previous build failed
        build_status_mock = MagicMock(return_value=BuildStatus.FAILED)
        assert not examinee(
                NotificationTriggeringPolicy.ONLY_FIRST,
                meta_vars={},
                determine_previous_build_status=build_status_mock,
                cfg_set=None,
        )
        assert examinee(
                NotificationTriggeringPolicy.ALWAYS,
                meta_vars={},
                determine_previous_build_status=build_status_mock,
                cfg_set=None,
        )
        assert not examinee(
                NotificationTriggeringPolicy.NEVER,
                meta_vars={},
                determine_previous_build_status=build_status_mock,
                cfg_set=None,
        )

    def test_cfg_from_callback(self):
        examinee = notification.cfg_from_callback

        callback_file = os.path.join(self.tmp_dir.name, 'call_me')
        with open(callback_file, 'w') as f:
            f.write('#!/usr/bin/env sh\necho "foo: 42">"${NOTIFY_CFG_OUT}"')
        os.chmod(callback_file, stat.S_IEXEC | stat.S_IREAD)

        assert examinee(
            repo_root=self.tmp_dir.name,
            callback_path=callback_file,
            effective_cfg_file='no-file-yet',
        ) == {'foo':42}



import logging
import typing

import cfg_mgmt
import cfg_mgmt.aws as cmaws
import cfg_mgmt.alicloud as cmali
import cfg_mgmt.azure as cma
import cfg_mgmt.btp_application_certificate as cmbac
import cfg_mgmt.btp_service_binding as cmb
import cfg_mgmt.gcp as cmg
import cfg_mgmt.github as cmgh
import cfg_mgmt.kubernetes as cmk
import cfg_mgmt.model as cmm
import model
import model.github
import oci.model as om


logger = logging.getLogger(__name__)


def delete_expired_secret(
    cfg_element: model.NamedModelElement,
    cfg_queue_entry: cmm.CfgQueueEntry,
    cfg_factory: model.ConfigFactory,
) -> tuple[model.NamedModelElement, bool]:
    '''Deletes the expired secret contained in the given cfg-queue entry, using the passed
    config element and config factory if necessary.

    Returns a pair containing a potentially changed cfg-element and a boolean indicating whether
    the deletion was successful.
    '''

    delete_func: typing.Callable[[model.NamedModelElement, str, cmm.CfgQueueEntry], None] = None

    if (type_name := cfg_queue_entry.target.type) == 'container_registry':
        if cfg_element.registry_type() in (
            om.OciRegistryType.GAR,
            om.OciRegistryType.GCR,
        ):
            delete_func = cmg.delete_config_secret
        else:
            f'{cfg_element.registry_type()} is not (yet) supported for automated deletion'
            return False

    elif type_name == 'gcp':
        delete_func = cmg.delete_config_secret

    elif type_name == 'github':
        delete_func = cmgh.delete_config_secret

    elif type_name == 'azure_service_principal':
        delete_func = cma.delete_config_secret

    elif type_name == 'btp_service_binding':
        delete_func = cmb.delete_config_secret

    elif type_name == 'btp_application_certificate':
        delete_func = cmbac.delete_config_secret

    elif type_name == 'aws':
        delete_func = cmaws.delete_config_secret

    elif type_name == 'alicloud':
        delete_func = cmali.delete_config_secret

    elif type_name == 'kubernetes':
        try:
            cmk.validate_for_rotation(cfg_element)
        # This can only happen if the kubernetes config was edited after the rotation
        # but before the removal of the expired secret
        except cmm.ValidationError as e:
            logger.warning(
                f"Cannot rotate cfg-type '{type_name}' with name '{cfg_element.name()}': {e}"
            )
            return None, None

        delete_func = cmk.delete_config_secret

    if not delete_func:
        logger.warning(
            f'{type_name} is not (yet) supported for automated deletion'
        )
        return None, False

    try:
        updated_cfg_element = delete_func(
            cfg_element=cfg_element,
            cfg_factory=cfg_factory,
            cfg_queue_entry=cfg_queue_entry,
        )
    except Exception as e:
        logger.error(
            f"error deleting secret for cfg-type '{type_name}' with name "
            f"'{cfg_element.name()}': {e}."
        )
        raise

    return updated_cfg_element, True


def rotate_cfg_element(
    cfg_element: model.NamedModelElement,
    cfg_factory: model.ConfigFactory,
) -> typing.Union[typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement], None]:
    '''Rotates the credentials contained in the given config-element, using only config from the
    given factory (if necessary)

    Returns a triple of (callable, dict, NamedModelElement) if rotation for the given config-
    element is supported:
    - The callable is a function that reverts the rotation again.
    - The dict contains the meta-information about the rotation.
    - The returned NamedModelElement is the updated version of the element that was passed in.

    If rotation for the given element is not supported, `None` will be returned.
    '''
    type_name = cfg_element._type_name

    update_secret_function: typing.Callable[
        [model.NamedModelElement, model.ConfigFactory],
        typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]
    ] = None
    rotation_validation_function: typing.Callable[[model.NamedModelElement], None] = None

    if type_name == 'container_registry':
        if cfg_element.registry_type() in (
            om.OciRegistryType.GAR,
            om.OciRegistryType.GCR,
        ):
            # service accounts require role "Service Account Key Admin"
            logger.info(f'rotating {cfg_element.name()} {type_name=}')
            update_secret_function = cmg.rotate_cfg_element
        else:
            logger.warning(
                f'{cfg_element.registry_type()} is not (yet) supported for automated rotation'
            )
            return None

    elif type_name == 'gcp':
        update_secret_function = cmg.rotate_cfg_element

    elif type_name == 'github':
        update_secret_function = cmgh.rotate_cfg_element

    elif type_name == 'azure_service_principal':
        update_secret_function = cma.rotate_cfg_element

    elif type_name == 'btp_service_binding':
        update_secret_function = cmb.rotate_cfg_element

    elif type_name == 'btp_application_certificate':
        update_secret_function = cmbac.rotate_cfg_element

    elif type_name == 'aws':
        rotation_validation_function = cmaws.validate_for_rotation
        update_secret_function = cmaws.rotate_cfg_element

    elif type_name == 'alicloud':
        rotation_validation_function = cmali.validate_for_rotation
        update_secret_function = cmali.rotate_cfg_element

    elif type_name == 'kubernetes':
        rotation_validation_function = cmk.validate_for_rotation
        update_secret_function = cmk.rotate_cfg_element

    if not update_secret_function:
        logger.warning(f'{type_name=} is not (yet) supported for automated rotation')
        return None

    try:
        if rotation_validation_function:
            rotation_validation_function(cfg_element)

        return update_secret_function(
            cfg_element=cfg_element,
            cfg_factory=cfg_factory,
        )

    except cmm.ValidationError as e:
        logger.warning(
            f"Cannot rotate cfg-type '{type_name}' with name '{cfg_element.name()}': {e}"
        )
        return None

    except Exception as e:
        logger.warning(f'an error occured whilst trying to update secret for {cfg_element=}: {e}')
        raise



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import base64
import json
import logging
import urllib.parse
import typing

import dacite

import ci.log
import ci.util
import oci.util
import oci.auth as oa
import oci.model as om

import model.base
from model.base import (
    BasicCredentials,
    NamedModelElement,
    ModelDefaultsMixin,
)

from ci.util import check_type


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


class ContainerRegistryConfig(NamedModelElement, ModelDefaultsMixin):
    '''
    Not intended to be instantiated by users of this module
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._apply_defaults(self.raw)

    def _defaults_dict(self):
        return {
            'privileges': oa.Privileges.READONLY.value,
        }

    def _optional_attributes(self):
        return {
            'image_reference_prefixes',
            'host',
            'email',
            'api_base_url',
            'registry_type',
            'rotation_cfg',
        }

    def _required_attributes(self):
        return {
            'username',
            'password',
        }

    def validate(self):
        super().validate()
        if (
            self.rotation_cfg()
            and self.registry_type() not in (
                om.OciRegistryType.GAR,
                om.OciRegistryType.GCR,
            )
        ):
            raise model.base.ModelValidationError(
                f'rotation_cfg not allowed for {self.registry_type()=}'
            )

    def client_email(self) -> str:
        return json.loads(self.password())['client_email']

    def private_key_id(self) -> str:
        return json.loads(self.password())['private_key_id']

    def registry_type(self):
        try:
            return om.OciRegistryType(self.raw.get('registry_type'))
        except ValueError:
            return om.OciRegistryType.UNKNOWN

    def api_base_url(self):
        return self.raw.get('api_base_url')

    def rotation_cfg(self) -> model.base.CfgElementReference:
        '''
        used to specify cfg-element to use for cross-rotation
        '''
        raw = self.raw.get('rotation_cfg')
        if raw:
            return dacite.from_dict(
                data_class=model.base.CfgElementReference,
                data=raw,
            )

        return None

    def credentials(self):
        # XXX handle different container registry types
        return GcrCredentials(self.raw)

    def has_service_account_credentials(self):
        return GcrCredentials(self.raw).has_service_account_credentials()

    def as_docker_auths(self):
        '''
        returns a representation of the credentials from this registry-cfg as "docker-auths",
        which can be used to populate a docker-cfg file ($HOME/.docker/config.json) below the
        `auths` attr
        '''
        auth_str = f'{self.credentials().username()}:{self.credentials().passwd()}'
        auth_str = base64.b64encode(auth_str.encode('utf-8')).decode('utf-8')

        hostnames = set()
        for image_prefix in self.image_reference_prefixes():
            if '://' not in image_prefix:
                image_prefix = f'x://{image_prefix}'
            hostnames.add(urllib.parse.urlparse(image_prefix).hostname)

        auths = {
            host: {'auth': auth_str} for host in hostnames
        }

        # ugly workaround for docker-hub, see:
        # https://github.com/GoogleContainerTools/kaniko/issues/1209
        # -> if there is a cfg for "docker-hub", implicitly patch-in the v1-flavoured cfg
        if not (docker_v1_prefix := 'https://index.docker.io/v1/') in auths:
            # try both w/ slash and w/o slash suffix
            def auth_for_prefix(prefix):
                if v := auths.get(prefix, None):
                    return v
                return auths.get(prefix + '/', None)

            for candidate in 'registry-1.docker.io', 'docker.io', 'index.docker.io':
                if auth_cfg := auth_for_prefix(prefix=candidate):
                    auths[docker_v1_prefix] = auth_cfg
                    break # use first (we have no way of knowing which one would be better..)

        logger.info(f'using cfg {self.name()=} for prefixes {list(auths.keys())}')

        return auths

    def privileges(self) -> oa.Privileges:
        return oa.Privileges(self.raw['privileges'])

    def image_reference_prefixes(self):
        prefixes = self.raw.get('image_reference_prefixes', ())
        if isinstance(prefixes, str):
            return [prefixes]
        return prefixes

    def image_ref_matches(
        self,
        image_reference: str,
        privileges: oa.Privileges=None,
    ):
        '''
        returns a boolean indicating whether a given container image reference matches any
        configured image reference prefixes (thus indicating this cfg might be adequate for
        retrieving or deploying the given container image using this cfg).

        If no image reference prefixes are configured, `False` is returned.
        '''
        check_type(image_reference, str)

        prefixes = self.image_reference_prefixes()
        if not prefixes:
            return False
        if privileges:
            # if privileges were specified, ours must be "great enough" (greater means more privs)
            if self.privileges() < privileges:
                return False

        for prefix in prefixes:
            if image_reference.startswith(prefix):
                return True
        return False


class GcrCredentials(BasicCredentials):
    '''
    Not intended to be instantiated by users of this module
    '''
    def _optional_attributes(self):
        return {
            'email',
            'host',
            'image_reference_prefixes',
        }

    def email(self):
        # used by KubernetesSecretsHelper's create_gcr_secret
        return self.raw.get('email')

    def host(self):
        # used in lss
        return self.raw.get('host')

    def has_service_account_credentials(self):
        '''
        heuristically (aka HACKY!!!) guesses whether the configured passwd _could_ be a
        GCP Service Account document
        '''
        try:
            json.loads(self.passwd())
            return True
        except json.decoder.JSONDecodeError:
            return False

    def service_account_credentials(self): # -> 'google.oauth2.service_account.Credentials':
        import google.oauth2.service_account
        return google.oauth2.service_account.Credentials.from_service_account_info(
            json.loads(self.passwd())
        )


def find_config(
    image_reference: typing.Union[str, om.OciImageReference],
    privileges:oa.Privileges=None,
    _normalised_image_reference=False,
    cfg_factory=None,
) -> typing.Optional[ContainerRegistryConfig]:
    image_reference = str(image_reference)
    if not cfg_factory:
        cfg_factory = ci.util.ctx().cfg_factory()

    if isinstance(image_reference, om.OciImageReference):
        image_reference = image_reference.normalised_image_reference()
        _normalised_image_reference = True

    matching_cfgs = sorted((
        cfg for cfg in
        cfg_factory._cfg_elements('container_registry')
        if cfg.image_ref_matches(image_reference, privileges=privileges)
        ),
        key=lambda c:c.privileges(),
    )

    if not matching_cfgs:
        # finally give up - did not match anything, even after normalisation
        if _normalised_image_reference:
            return None
        else:
            return find_config(
                image_reference=oci.util.normalise_image_reference(image_reference=image_reference),
                privileges=privileges,
                _normalised_image_reference=True,
                cfg_factory=cfg_factory,
            )

    # return first match (because they are sorted, this will be the one with least privileges)
    registry_cfg = matching_cfgs[0]

    return registry_cfg



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import ensure
import functools

import ci.log
import ci.util
import concourse.client.api
import ctx
import model.concourse


def lookup_cc_team_cfg(
    concourse_cfg,
    cfg_set,
    team_name,
) -> model.concourse.ConcourseTeamConfig:
    for cc_team_cfg in cfg_set._cfg_elements('concourse_team_cfg'):
        if cc_team_cfg.team_name() != team_name:
            continue
        if concourse_cfg.concourse_endpoint_name() != cc_team_cfg.concourse_endpoint_name():
            continue

        return cc_team_cfg
    raise KeyError(f'No concourse team config for team name {team_name} found')


@ensure.ensure_annotations
def client_from_parameters(
    base_url: str,
    password: str,
    team_name: str,
    username: str,
    verify_ssl: bool = True,
    concourse_api_version=None,
) -> concourse.client.api.ConcourseApiBase:
    """
    returns a concourse-client w/ the credentials valid for the current execution environment.
    The returned client is authorised to perform operations in the same concourse-team as the
    credentials provided calling this function.
    """

    concourse_api = concourse.client.api.ConcourseApiFactory.create_api(
        base_url=base_url,
        team_name=team_name,
        verify_ssl=verify_ssl,
        concourse_api_version=concourse_api_version,
    )

    concourse_api.login(
        username=username,
        passwd=password,
    )
    return concourse_api


@functools.lru_cache()
@ensure.ensure_annotations
def client_from_cfg_name(
    concourse_cfg_name: str,
    team_name: str,
    cfg_factory=None,
):
    '''Create a Concourse API client for the team with the given name on the given concourse instance

    As team names are not necessarily unique across all our Concourse instances, this function
    will perform a lookup using the given config-factory/-set. If no config-factory or -set is given,
    all config that is available by default will be considered.
    An error will be raised if no team with the requested name can be found for the given Concourse
    instance in the config.
    '''
    if not cfg_factory:
        cfg_factory = ci.util.ctx().cfg_factory()

    concourse_cfg = cfg_factory.concourse(concourse_cfg_name)

    concourse_team_config = lookup_cc_team_cfg(
        concourse_cfg=concourse_cfg,
        cfg_set=cfg_factory,
        team_name=team_name,
    )
    concourse_endpoint = cfg_factory.concourse_endpoint(
        concourse_team_config.concourse_endpoint_name()
    )
    return client_from_parameters(
        base_url=concourse_endpoint.base_url(),
        password=concourse_team_config.password(),
        team_name=team_name,
        username=concourse_team_config.username(),
    )


def client_from_env(
    team_name: str=None,
    cfg_set=None
):
    if not cfg_set:
        cfg_set = ctx.cfg_set()

    if not team_name:
        team_name = ci.util.check_env('CONCOURSE_CURRENT_TEAM')

    concourse_team_config = lookup_cc_team_cfg(
        concourse_cfg=cfg_set.concourse(),
        cfg_set=cfg_set,
        team_name=team_name,
    )
    concourse_endpoint = cfg_set.concourse_endpoint(
        concourse_team_config.concourse_endpoint_name()
    )
    return client_from_parameters(
        base_url=concourse_endpoint.base_url(),
        password=concourse_team_config.password(),
        team_name=team_name,
        username=concourse_team_config.username(),
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import enum

from model.base import (
    ModelBase,
    NamedModelElement,
)
from model.proxy import DockerImageConfig


class SigningAlgorithm(enum.StrEnum):
    RSASSA_PSS = 'rsassa-pss'
    RSASSA_PKCS1_V1_5 = 'rsassa-pkcs1-v1_5'


class SigningServerEndpoint(NamedModelElement):
    def url(self) -> str:
        return self.raw.get('url')


class SigningServerIngressConfig(ModelBase):

    def enabled(self) -> bool:
        return self.raw['enabled']

    def auth(self):
        return self.raw.get('auth')

    def _required_attributes(self):
        return {
            'enabled'
        }


class SigningServerConfig(NamedModelElement):

    def image_config(self) -> DockerImageConfig:
        return DockerImageConfig(self.raw['image'])

    def namespace(self) -> str:
        return self.raw['namespace']

    def log_config(self) -> dict:
        return self.raw['log']

    def replica_count(self) -> int:
        return self.raw['replica_count']

    def image_pull_secret_name(self) -> str:
        return self.raw['image_pull_secret_name']

    def private_key_secret_name(self) -> str:
        return self.raw['private_key_secret_name']

    def private_key(self) -> str:
        return self.raw['private_key']

    def certificate_configmap_name(self) -> str:
        return self.raw['certificate_configmap_name']

    def certificate(self):
        return self.raw['certificate']

    def ca_certificates_configmap_name(self) -> str:
        return self.raw['ca_certificates_configmap_name']

    def ca_certificates(self):
        return self.raw['ca_certificates']

    def max_body_size(self) -> int:
        return self.raw['max_body_size']

    def disable_auth(self) -> bool:
        return self.raw['disable_auth']

    def host(self) -> str:
        return self.raw['host']

    def cosign_repository(self) -> str:
        return self.raw['cosign_repository']

    def ingress_config(self) -> SigningServerIngressConfig:
        return SigningServerIngressConfig(self.raw['ingress'])

    def algorithm(self) -> SigningAlgorithm:
        return SigningAlgorithm(self.raw.get('algorithm', SigningAlgorithm.RSASSA_PSS))

    def signature_name(self) -> str:
        return self.raw['signature_name']



import datetime

import dacite
import dateutil.parser

# keep import short for convenience
from dataclasses import (
    dataclass,
    field,
)

from enum import (
    Enum,
    IntEnum,
)
from typing import(
    List,
    Optional,
)


class Severity(IntEnum):
    CRITICAL = 5
    HIGH = 4
    MEDIUM = 3
    LOW = 2
    MINIMAL = 1
    SEVERITY_UNSPECIFIED = 0


class OccurrenceKind(Enum):
    DISCOVERY = 'DISCOVERY'
    VULNERABILITY = 'VULNERABILITY'


class PackageVersionKind(Enum):
    NORMAL = 'NORMAL'
    MAXIMUM = 'MAXIMUM'


class ContinuousAnalysis(Enum):
    INACTIVE = 'INACTIVE'
    ACTIVE = 'ACTIVE'
    CONTINUOUS_ANALYSIS_UNSPECIFIED =  'CONTINUOUS_ANALYSIS_UNSPECIFIED'


class AnalysisStatus(Enum):
    FINISHED_SUCCESS = 'FINISHED_SUCCESS'
    PENDING = 'PENDING'
    SCANNING = 'SCANNING'
    FINISHED_FAILED = 'FINISHED_FAILED'
    FINISHED_UNSUPPORTED = 'FINISHED_UNSUPPORTED'
    ANALYSIS_STATUS_UNSPECIFIED = 'ANALYSIS_STATUS_UNSPECIFIED'


@dataclass
class PackageVersion:
    kind: PackageVersionKind
    name: Optional[str]
    revision: Optional[str]
    epoch: Optional[int]
    fullName: Optional[str]


@dataclass
class PackageIssue:
    affectedCpeUri: str
    affectedPackage: str
    affectedVersion: PackageVersion
    fixedCpeUri: str
    fixedPackage: str
    fixedVersion: PackageVersion
    fixAvailable: bool = False


@dataclass
class Discovery:
    analysisStatus: Optional[AnalysisStatus]
    analysisStatusError: Optional[dict]
    continuousAnalysis: Optional[ContinuousAnalysis]


@dataclass
class Vulnerability:
    packageIssue: List[PackageIssue]
    severity: Optional[Severity]
    relatedUrls: Optional[List[dict]]
    longDescription: Optional[str]
    shortDescription: Optional[str]
    effectiveSeverity: Severity = Severity.SEVERITY_UNSPECIFIED
    fixAvailable: bool = False
    cvssScore: float = 0.0 # In line with google's UI display


@dataclass
class Resource:
    uri: str
    contentHash: Optional[dict]
    name: Optional[str]


@dataclass
class Occurrence:
    createTime: datetime.datetime
    kind: OccurrenceKind
    name: str
    noteName: str
    resourceUri: str
    updateTime: datetime.datetime
    vulnerability: Optional[Vulnerability]
    discovery: Optional[Discovery]


@dataclass
class ListOccurrencesResponse:
    nextPageToken: Optional[str]
    occurrences: Optional[List[Occurrence]] = field(default_factory=list)

    @staticmethod
    def parse(response):
        return dacite.from_dict(
            data_class=ListOccurrencesResponse,
            data=response,
            config=dacite.Config(
                type_hooks={
                    Severity: lambda x: Severity[x],
                    datetime.datetime: lambda x: dateutil.parser.isoparse(x),
                },
                cast=[Enum],
            ),
        )



import copy
import dataclasses
import datetime
import logging
import typing

import boto3
import dacite

import cfg_mgmt
import ci.log
import model
import model.aws
import model.base

from cfg_mgmt.model import (
    CfgQueueEntry,
    ValidationError,
)


logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


@dataclasses.dataclass
class AwsAccessKey:
    UserName: str
    AccessKeyId: str
    Status: str
    SecretAccessKey: str
    CreateDate: datetime.datetime


@dataclasses.dataclass
class AccessKeyMetadata:
    UserName: str
    AccessKeyId: str
    Status: str
    CreateDate: datetime.datetime


@dataclasses.dataclass
class CreateAccessKeyResponse:
    AccessKey: AwsAccessKey


@dataclasses.dataclass
class ListAccessKeysResponse:
    AccessKeyMetadata: typing.List[AccessKeyMetadata]
    IsTruncated: bool
    Marker: typing.Optional[str] # only present when IsTruncated is True


def rotation_cfg_or_none(
    aws_cfg: model.aws.AwsProfile,
    cfg_factory: model.ConfigFactory,
):
    if (rotation_cfg_reference := aws_cfg.rotation_cfg()):
        return cfg_factory.resolve_cfg_element_reference(
            cfg_element_reference=rotation_cfg_reference,
        )

    return None


def rotate_cfg_element(
    cfg_element: model.aws.AwsProfile,
    cfg_factory: model.ConfigFactory,
) ->  typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:
    rotation_cfg = rotation_cfg_or_none(
        aws_cfg=cfg_element,
        cfg_factory=cfg_factory,
    ) or cfg_element

    logger.info(f'using {rotation_cfg.name()=} for rotation')

    iam_user_name = cfg_element.iam_user_name()

    iam_client = boto3.client(
        'iam',
        aws_access_key_id=rotation_cfg.access_key_id(),
        aws_secret_access_key=rotation_cfg.secret_access_key(),
        region_name=rotation_cfg.region(),
    )

    if iam_user_name:
        data = iam_client.create_access_key(UserName=iam_user_name)

    else:
        data = iam_client.create_access_key()

    response: CreateAccessKeyResponse = dacite.from_dict(
        data_class=CreateAccessKeyResponse,
        data=data,
    )
    access_key = response.AccessKey

    raw_cfg = copy.deepcopy(cfg_element.raw)
    new_element = model.aws.AwsProfile(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )
    new_element.raw['access_key_id'] = access_key.AccessKeyId
    new_element.raw['secret_access_key'] = access_key.SecretAccessKey

    def revert_function():
        if iam_user_name:
            iam_client.delete_access_key(
                AccessKeyId=access_key.AccessKeyId,
                UserName=iam_user_name,
            )
        else:
            iam_client.delete_access_key(AccessKeyId=access_key.AccessKeyId)

    secret_id = {'accessKeyId': cfg_element.access_key_id()}

    return revert_function, secret_id, new_element


def delete_config_secret(
    cfg_element: model.aws.AwsProfile,
    cfg_factory: model.ConfigFactory,
    cfg_queue_entry: CfgQueueEntry,
) -> model.aws.AwsProfile | None:
    rotation_cfg = rotation_cfg_or_none(
        aws_cfg=cfg_element,
        cfg_factory=cfg_factory,
    ) or cfg_element

    logger.info(f'using {rotation_cfg.name()=} for deletion')

    iam_user_name = cfg_element.iam_user_name()

    iam_client = boto3.client(
        'iam',
        aws_access_key_id=rotation_cfg.access_key_id(),
        aws_secret_access_key=rotation_cfg.secret_access_key(),
        region_name=rotation_cfg.region(),
    )
    access_key_id = cfg_queue_entry.secretId['accessKeyId']

    if iam_user_name:
        iam_client.delete_access_key(
            UserName=iam_user_name,
            AccessKeyId=access_key_id,
        )

    else:
        iam_client.delete_access_key(AccessKeyId=access_key_id)

    return None


def validate_for_rotation(
    cfg_element: model.aws.AwsProfile,
):
    cfg_element.validate()

    access_key_id = cfg_element.access_key_id()
    iam_client = boto3.client(
        'iam',
        aws_access_key_id=access_key_id,
        aws_secret_access_key=cfg_element.secret_access_key(),
        region_name=cfg_element.region(),
    )
    response: ListAccessKeysResponse = dacite.from_dict(
        data_class=ListAccessKeysResponse,
        data=iam_client.list_access_keys(),
    )
    key_metadata = response.AccessKeyMetadata

    if len(key_metadata) == 2:
        raise ValidationError(
            'There are already two keys present in AWS for Access Key '
            f"'{access_key_id}'. Will not attempt rotation."
        )



import tempfile
import unittest
import os
import pathlib

import test_utils

from concourse.steps import step_def
import concourse.model.traits.version as version_trait


class VersionStepTest(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.version_outdir = pathlib.Path(self.tmp_dir.name, 'managed-version')
        self.version_outdir.mkdir()

        self.version_trait = version_trait.VersionTrait(
            name='version',
            variant_name='don\'t_care',
            raw_dict={
                'versionfile': 'version',
                'preprocess': 'finalize',
            },
        )
        self.version_trait_transformer = self.version_trait.transformer()
        self.render_step = step_def('version')

        self.main_repo = test_utils.repository()

        repo_dir = pathlib.Path(self.tmp_dir.name, self.main_repo.resource_name())
        repo_dir.mkdir()

        # create fake git dir
        os.mkdir(repo_dir.joinpath('.git'))
        repo_dir.joinpath('.git', 'HEAD').touch()

        self.version_file = repo_dir.joinpath('version')
        self.version_file.write_text('1.2.3-xxx')

        self.job_variant = test_utils.job(self.main_repo)

        self.job_variant._traits_dict = {'version': self.version_trait}

        for step in self.version_trait_transformer.inject_steps():
            self.job_variant._steps_dict[step.name] = step

        self.old_cwd = os.getcwd()

    def tearDown(self):
        self.tmp_dir.cleanup()
        os.chdir(self.old_cwd)

    def test_render_and_compile(self):
        # as a smoke-test, just try to render
        step_snippet = self.render_step(
            job_step=self.job_variant.step('version'),
            job_variant=self.job_variant,
            indent=0
        )

        # try to compile (-> basic syntax check)
        return compile(step_snippet, 'version', 'exec')

    def test_smoke(self):
        compiled = self.test_render_and_compile()

        os.chdir(self.tmp_dir.name)

        eval(compiled)

        # check that version preprocessing was actually done
        effective_version = pathlib.Path(self.tmp_dir.name, 'managed-version', 'version')
        self.assertEqual(effective_version.read_text(), '1.2.3')



import functools
import logging
import urllib

import google.oauth2.service_account as service_account
import googleapiclient.discovery
import google.cloud.storage

import ccc.oci
import model.container_registry
import model.gcp

import ci.util
import oci.util
from .grafeas_model import (
    AnalysisStatus,
    ContinuousAnalysis,
    ListOccurrencesResponse,
    Severity,
)


def logger():
    return logging.getLogger(__name__)


logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)


def _to_gcp_cfg(gcp_cfg: str):
    if isinstance(gcp_cfg, str):
        cfg_factory = ci.util.ctx().cfg_factory()
        gcp_cfg = cfg_factory.gcp(gcp_cfg)
    return gcp_cfg


def credentials(gcp_cfg: str):
    gcp_cfg = _to_gcp_cfg(gcp_cfg=gcp_cfg)

    credentials = service_account.Credentials.from_service_account_info(
        gcp_cfg.service_account_key(),
    )

    return credentials


def authenticated_build_func(gcp_cfg: str):
    creds = credentials(gcp_cfg=gcp_cfg)

    return functools.partial(googleapiclient.discovery.build, credentials=creds)


def cloud_storage_client(gcp_cfg: str, *args, **kwargs):
    gcp_cfg = _to_gcp_cfg(gcp_cfg=gcp_cfg)
    creds = credentials(gcp_cfg=gcp_cfg)

    return google.cloud.storage.Client(
        project=gcp_cfg.project(),
        credentials=creds,
        *args,
        **kwargs,
    )


CONTAINERANALYSIS_DEFAULT_AUTH_SCOPES = ('https://www.googleapis.com/auth/cloud-platform',)


class VulnerabilitiesRetrievalFailed(RuntimeError):
    pass


def qualified_service_account_name(
    service_account_name: str,
    project_id: str = '-',
) -> str:
    return f'projects/{project_id}/serviceAccounts/{service_account_name}'


def qualified_service_account_key_name(
    service_account_name: str,
    key_name: str,
    project_id: str = '-',
) -> str:
    base_name = qualified_service_account_name(
        service_account_name=service_account_name,
        project_id=project_id,
    )
    return f'{base_name}/keys/{key_name}'


def create_iam_client(
    cfg_element: model.container_registry.ContainerRegistryConfig | model.gcp.GcpServiceAccount,
) -> googleapiclient.discovery.Resource:
    if isinstance(cfg_element, model.container_registry.ContainerRegistryConfig):
        credentials = cfg_element.credentials().service_account_credentials()
    elif isinstance(cfg_element, model.gcp.GcpServiceAccount):
        credentials = cfg_element.service_account_credentials()
    else:
        raise NotImplementedError

    return googleapiclient.discovery.build(
        serviceName='iam',
        version='v1',
        credentials=credentials,
    )


class GrafeasClient:
    def __init__(
        self,
        container_registry_config,
        scopes=CONTAINERANALYSIS_DEFAULT_AUTH_SCOPES,
    ):
        credentials = container_registry_config.credentials().service_account_credentials()
        scoped_credentials = credentials.with_scopes(scopes)
        self._api_client =  googleapiclient.discovery.build(
            'containeranalysis',
            'v1',
            credentials=scoped_credentials,
            static_discovery=False,
        )

    @staticmethod
    def for_image(
        image_reference: str,
        scopes=CONTAINERANALYSIS_DEFAULT_AUTH_SCOPES,
    ):
        '''Convenience function for client creation

        NOTE: Will determine credentials to use from image reference. The created client's methods
        will *not* work for other images to which the determined credentials have no access to.
        '''
        image_reference = oci.util.normalise_image_reference(image_reference)
        registry_config = model.container_registry.find_config(image_reference=image_reference)

        if not registry_config:
            raise VulnerabilitiesRetrievalFailed(f'no registry-cfg found for: {image_reference}')
        if not registry_config.has_service_account_credentials():
            raise VulnerabilitiesRetrievalFailed(
                f'no gcr-cfg {registry_config.name()} {image_reference}'
            )

        return GrafeasClient(
            container_registry_config=registry_config,
            scopes=scopes,
        )

    def _list_occurrences(self, project, filter_expression):
        projects = self._api_client.projects() # noqa; pylint: disable=E1101
        occurrences = projects.occurrences()

        request = occurrences.list(
            parent=project,
            filter=filter_expression,
        )
        while request is not None:
            response = request.execute()
            parsed_response = ListOccurrencesResponse.parse(response)
            yield from parsed_response.occurrences

            request = occurrences.list_next(request, response)

    def _parse_gcr_parameters(self, image_reference):
        project_name = urllib.parse.urlparse(image_reference).path.split('/')[1]
        oci_client = ccc.oci.oci_client()
        try:
            hash_reference = oci_client.to_digest_hash(image_reference=image_reference)
        except Exception as e:
            raise VulnerabilitiesRetrievalFailed(e) from e

        return project_name, hash_reference

    def retrieve_vulnerabilities(
        self,
        image_reference: str,
    ):
        # XXX / HACK: assuming we always handle GCRs (we should rather check!), the first URL path
        # element is the GCR project name
        project_name, hash_reference = self._parse_gcr_parameters(image_reference)

        logger().info(f'retrieving vulnerabilites for {project_name=} / {hash_reference=}')

        filter_str = f'resourceUrl = "https://{hash_reference}" AND kind="VULNERABILITY"'

        occurrences = self._list_occurrences(
            project=f'projects/{project_name}',
            filter_expression=filter_str,
        )

        yield from occurrences

    def filter_vulnerabilities(
        self,
        image_reference: str,
        cvss_threshold: float=7.0,
        effective_severity_threshold: Severity=Severity.SEVERITY_UNSPECIFIED
    ):
        try:
            vulnerabilityOccurrences = self.retrieve_vulnerabilities(image_reference=image_reference)
        except Exception as e:
            raise VulnerabilitiesRetrievalFailed(e)

        for occurrence in vulnerabilityOccurrences:
            vulnerability = occurrence.vulnerability

            if vulnerability.cvssScore < cvss_threshold:
                continue

            if (
                vulnerability.effectiveSeverity is not Severity.SEVERITY_UNSPECIFIED
                and vulnerability.effectiveSeverity < effective_severity_threshold
            ):
                continue

            # return everything that was not filtered out
            yield occurrence

    def scan_available(
        self,
        image_reference: str,
    ):
        image_reference = oci.util.normalise_image_reference(image_reference)

        # XXX / HACK: assuming we always handle GCRs (we should rather check!), the first URL path
        # element is the GCR project name
        project_name, hash_reference = self._parse_gcr_parameters(image_reference)

        filter_str = f'resourceUrl = "https://{hash_reference}" AND kind="DISCOVERY"'

        discoveries = list(self._list_occurrences(
            project=f'projects/{project_name}',
            filter_expression=filter_str,
        ))

        if (discovery_count := len(discoveries)) == 0:
            logger().warning(f'found no discovery-info for {image_reference=}')
            return False
        elif discovery_count > 1:
            # use latest
            timestamp = -1
            candidate = None
            for d in discoveries:
                timestamp = max(timestamp, d.updateTime.timestamp())
                if timestamp == d.updateTime.timestamp():
                    candidate = d
        else:
            candidate = discoveries[0]

        discovery = candidate.discovery

        discovery_status = discovery.analysisStatus
        continuous_analysis = discovery.continuousAnalysis

        # XXX hard-code we require continuous scanning to be enabled
        if not continuous_analysis or continuous_analysis is not ContinuousAnalysis.ACTIVE:
            logger().warning(f'{continuous_analysis=} for {image_reference=}')
            return False

        if not discovery_status or discovery_status is not AnalysisStatus.FINISHED_SUCCESS:
            logger().warning(f'{discovery_status=} for {image_reference=}')
            return False

        return True # finally



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
)


class PyPiCfg(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def credentials(self):
        return PyPiCredentials(self.raw.get('credentials'))


class PyPiCredentials(BasicCredentials):
    pass



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import datetime
import enum
import functools
import logging
import traceback
import urllib.parse

import cachecontrol
import deprecated
import ocm
import github3
import github3.github
import github3.session

import ci.util
import github.util
import http_requests
import model
import model.github
import model.base

logger = logging.getLogger(__name__)


class SessionAdapter(enum.Enum):
    NONE = None
    RETRY = 'retry'
    CACHE = 'cache'


def github_api_ctor(
    github_cfg: model.github.GithubConfig,
    github_username: str,
    verify_ssl: bool=True,
    session_adapter: SessionAdapter=SessionAdapter.RETRY,
    cfg_factory: model.ConfigFactory=None,
):
    '''returns the appropriate github3.GitHub constructor for the given github URL

    In case github_url does not refer to github.com, the c'tor for GithubEnterprise is
    returned with the url argument preset, thus disburdening users to differentiate
    between github.com and non-github.com cases.

    github-api requests will be logged to Elasticsearch if a default Elasticsearch config is found.
    '''
    github_url = github_cfg.http_url()

    parsed = urllib.parse.urlparse(github_url)
    if parsed.scheme:
        hostname = parsed.hostname
    else:
        raise ValueError('failed to parse url: ' + str(github_url))

    session = github3.session.GitHubSession()
    original_request = session.request

    def intercepted_request(
        method: str,
        url: str,
        **kwargs,
    ):
        req = original_request(method, url, **kwargs)

        try:
            es_client.store_document(
                index='github_request',
                body={
                    'method': method,
                    'url': url,
                    'data': kwargs,
                    'creation_date': datetime.datetime.now().isoformat(),
                    'github_cfg': github_cfg.name(),
                    'github_hostname': hostname,
                    'github_username': github_username,
                },
            )
        except:
            logger.debug('unable to log github api request to Elasticsearch, '
                'will disable logging for future requests')
            session.request = original_request

        return req

    try:
        import ccc.elasticsearch
        es_client = ccc.elasticsearch.default_client_if_available(cfg_factory)
        if es_client:
            logger.debug('logging github api requests to elasticsearch')
            session.request = intercepted_request

    except Exception as e:
        logger.debug(e)
        logger.debug('unable to create elasticsearch client, will not log github-api requests')

    session_adapter = SessionAdapter(session_adapter)
    if session_adapter is SessionAdapter.NONE or not session_adapter:
        pass
    elif session_adapter is SessionAdapter.RETRY:
        session = http_requests.mount_default_adapter(
            session=session,
            flags=http_requests.AdapterFlag.RETRY,
            max_pool_size=16, # increase with care, might cause github api "secondary-rate-limit"
        )
    elif session_adapter is SessionAdapter.CACHE:
        session = cachecontrol.CacheControl(
            session,
            cache_etags=True,
        )
    else:
        raise NotImplementedError

    if hostname.lower() == 'github.com':
        return functools.partial(
            github3.github.GitHub,
            session=session,
        )
    else:
        return functools.partial(
            github3.github.GitHubEnterprise,
            url=github_url,
            verify=verify_ssl,
            session=session,
        )


def repo_helper(
    host: str,
    org: str,
    repo: str,
    branch: str='master',
    session_adapter: SessionAdapter=SessionAdapter.RETRY,
):
    api = github_api(
        github_cfg=github_cfg_for_repo_url(repo_url=ci.util.urljoin(host, org, repo)),
        session_adapter=session_adapter,
    )

    return github.util.GitHubRepositoryHelper(
        owner=org,
        name=repo,
        github_api=api,
        default_branch=branch,
    )


def pr_helper(
    host: str,
    org: str,
    repo: str,
    session_adapter: SessionAdapter=SessionAdapter.RETRY,
):
    api = github_api(
        github_cfg=github_cfg_for_repo_url(repo_url=ci.util.urljoin(host, org, repo)),
        session_adapter=session_adapter,
    )

    return github.util.PullRequestUtil(
        owner=org,
        name=repo,
        github_api=api,
    )


# XXX remove this alias again
github_repo_helper = repo_helper


def github_api(
    github_cfg: model.github.GithubConfig=None,
    repo_url: str=None,
    session_adapter: SessionAdapter=SessionAdapter.RETRY,
    cfg_factory=None,
    username: str | None=None,
):
    if not (bool(github_cfg) ^ bool(repo_url)):
        raise ValueError('exactly one of github_cfg, repo_url must be passed')

    if not cfg_factory:
        try:
            cfg_factory = ci.util.ctx().cfg_factory()
        except Exception as e:
            logger.warning(f'error trying to retrieve {repo_url=} {github_cfg=}: {e}')
            raise

    if isinstance(github_cfg, str):
        github_cfg = cfg_factory().github(github_cfg)

    if repo_url:
        github_cfg = github_cfg_for_repo_url(
            repo_url=repo_url,
            cfg_factory=cfg_factory,
        )

    if username:
        github_credentials = github_cfg.credentials(username)
    else:
        github_credentials = github_cfg.credentials_with_most_remaining_quota()

    github_auth_token = github_credentials.auth_token()
    github_username = github_credentials.username()

    verify_ssl = github_cfg.tls_validation()

    github_ctor = github_api_ctor(
        github_cfg=github_cfg,
        github_username=github_username,
        verify_ssl=verify_ssl,
        session_adapter=session_adapter,
        cfg_factory=cfg_factory,
    )
    github_api = github_ctor(
        token=github_auth_token,
    )

    if not github_api:
        ci.util.fail(f'Could not connect to GitHub-instance {github_cfg.http_url()}')

    if not 'github.com' in github_cfg.api_url():
        github_api._github_url = github_cfg.api_url()

    return github_api


@functools.lru_cache()
def github_cfg_for_repo_url(
    repo_url: str | urllib.parse.ParseResult=None,
    api_url: str=None,
    cfg_factory=None,
    require_labels: tuple[str]=('ci',), # XXX unhardcode label
    github_cfgs: tuple[model.github.GithubConfig]=(),
) -> model.github.GithubConfig | None:
    if not (bool(repo_url) ^ bool(api_url)):
        raise ValueError('exactly one of `repo_url` or `api_url` must be passed')

    if isinstance(repo_url, urllib.parse.ParseResult):
        repo_url = repo_url.geturl()

    if not github_cfgs:
        if not cfg_factory:
            cfg_factory = ci.util.ctx().cfg_factory()

        github_cfgs = cfg_factory._cfg_elements(cfg_type_name='github')

    matching_cfgs = []
    for github_cfg in github_cfgs:
        if require_labels:
            missing_labels = set(require_labels) - set(github_cfg.purpose_labels())
            if missing_labels:
                # if not all required labels are present skip this element
                continue

        if (
            (repo_url and github_cfg.matches_repo_url(repo_url=repo_url))
            or (api_url and github_cfg.matches_api_url(api_url=api_url))
        ):
            matching_cfgs.append(github_cfg)

    # prefer config with most configured repo urls
    matching_cfgs = sorted(matching_cfgs, key=lambda config: len(config.repo_urls()))
    url = repo_url or api_url
    if len(matching_cfgs) == 0:
        raise model.base.ConfigElementNotFoundError(f'No github cfg found for {url=}')

    gh_cfg = matching_cfgs[-1]
    logger.debug(f'using {gh_cfg.name()=} for {url=}')
    return gh_cfg


@deprecated.deprecated()
@functools.lru_cache()
def github_cfg_for_hostname(
    host_name,
    cfg_factory=None,
    require_labels=('ci',), # XXX unhardcode label
):
    ci.util.not_none(host_name)
    if not cfg_factory:
        ctx = ci.util.ctx()
        cfg_factory = ctx.cfg_factory()

    if isinstance(require_labels, str):
        require_labels = tuple(require_labels)

    def has_required_labels(github_cfg):
        if not require_labels:
            return True

        for required_label in require_labels:
            if required_label not in github_cfg.purpose_labels():
                return False
        return True

    for github_cfg in filter(has_required_labels, cfg_factory._cfg_elements(cfg_type_name='github')):
        if github_cfg.matches_hostname(host_name=host_name):
            return github_cfg

    raise RuntimeError(f'no github_cfg for {host_name} with {require_labels}')


def log_stack_trace_information_hook(resp, *args, **kwargs):
    '''
    This function stores the current stacktrace in elastic search.
    It must not return anything, otherwise the return value is assumed to replace the response
    '''
    if not ci.util._running_on_ci():
        return # early exit if not running in ci job

    import ccc.elasticsearch
    config_set_name = ci.util.check_env('CONCOURSE_CURRENT_CFG')
    try:
        els_index = 'github_access_stacktrace'
        try:
            config_set = ci.util.ctx().cfg_factory().cfg_set(config_set_name)
        except KeyError:
            # do nothing: external concourse does not have config set 'internal_active'
            return
        elastic_cfg = config_set.elasticsearch()

        now = datetime.datetime.utcnow()
        json_body = {
            'date': now.isoformat(),
            'url': resp.url,
            'req_method': resp.request.method,
            'stacktrace': traceback.format_stack()
        }

        elastic_client = ccc.elasticsearch.from_cfg(elasticsearch_cfg=elastic_cfg)
        elastic_client.store_document(
            index=els_index,
            body=json_body
        )

    except Exception as e:
        ci.util.info(f'Could not log stack trace information: {e}')


def github_api_from_gh_access(
    access: ocm.GithubAccess,
) -> github3.github.GitHub | github3.github.GitHubEnterprise:
    if access.type is not ocm.AccessType.GITHUB:
        raise ValueError(f'{access=}')

    github_cfg = github_cfg_for_repo_url(repo_url=access.repoUrl)
    return github_api(github_cfg=github_cfg)
